# -*- coding: utf-8 -*-
import itertools
import struct

import pytest
from mock import Mock, call, patch

from nameko_grpc.errors import GrpcError
from nameko_grpc.streams import (
    STREAM_END,
    ByteBuffer,
    ReceiveStream,
    SendStream,
    StreamBase,
)


class TestByteBuffer:
    def test_peek(self):
        buffer = ByteBuffer()
        buffer.write(b"abcdefghi")

        assert buffer.peek(slice(0, 1)) == b"a"
        assert buffer.peek(slice(3, 6)) == b"def"
        assert buffer.peek(slice(-2, -1)) == b"h"

        assert buffer.read() == b"abcdefghi"

    def test_peek_all(self):
        buffer = ByteBuffer()
        buffer.write(b"abcdefghi")

        assert buffer.peek() == b"abcdefghi"
        assert buffer.read() == b"abcdefghi"

    def test_discard(self):
        buffer = ByteBuffer()
        buffer.write(b"abcdefghi")

        assert buffer.discard(3) is None
        assert buffer.read() == b"defghi"

    def test_discard_all(self):
        buffer = ByteBuffer()
        buffer.write(b"abcdefghi")

        assert buffer.discard() is None
        assert buffer.read() == b""

    def test_read(self):
        buffer = ByteBuffer()
        buffer.write(b"abcdefghi")

        assert buffer.read(3) == b"abc"
        assert buffer.read() == b"defghi"

    def test_read_all(self):
        buffer = ByteBuffer()
        buffer.write(b"abcdefghi")

        assert buffer.read() == b"abcdefghi"
        assert buffer.read() == b""

    def test_write(self):
        buffer = ByteBuffer()

        buffer.write(b"abc")
        assert buffer.peek() == b"abc"
        buffer.write(b"def")
        assert buffer.peek() == b"abcdef"

    def test_empty(self):
        buffer = ByteBuffer()

        assert buffer.empty() is True
        buffer.write(b"abc")
        assert buffer.empty() is False
        buffer.discard()
        assert buffer.empty() is True

    def test_len(self):
        buffer = ByteBuffer()

        assert len(buffer) == 0
        buffer.write(b"abc")
        assert len(buffer) == 3


class TestStreamBase:
    def test_exhausted(self):
        stream = StreamBase(1)
        stream.buffer.write(b"abc")
        assert not stream.exhausted

        stream.close()
        assert stream.closed
        assert not stream.exhausted

        stream.queue.get()
        assert stream.queue.empty()
        assert not stream.exhausted

        stream.buffer.discard()
        assert stream.buffer.empty()
        assert stream.exhausted

    def test_close(self):
        stream = StreamBase(1)
        stream.close()

        assert stream.closed
        assert stream.queue.get() == STREAM_END

    def test_close_with_error(self):
        stream = StreamBase(1)
        error = GrpcError("boom", "details")
        stream.close(error)

        assert stream.closed
        assert stream.queue.get() == error

    def test_close_with_non_error(self):
        stream = StreamBase(1)
        error = Exception("boom")

        with pytest.raises(AssertionError):
            stream.close(error)


class TestReceiveStream:
    def test_write_to_closed_stream(self):
        stream = ReceiveStream(1)

        assert stream.buffer.empty()
        stream.close()
        stream.write(b"\x00\x00\x00")
        assert stream.buffer.empty()

    def test_write_less_bytes_than_header(self):
        stream = ReceiveStream(1)
        stream.write(b"\x00\x00\x00")

        assert stream.queue.empty()
        assert stream.buffer.peek() == b"\x00\x00\x00"

    def test_write_less_bytes_than_one_message(self):
        stream = ReceiveStream(1)
        stream.write(b"\x00\x00\x00\x01\x00\xff\xff\xff")

        assert stream.queue.empty()
        assert stream.buffer.peek() == b"\x00\x00\x00\x01\x00\xff\xff\xff"

    def test_write_more_bytes_than_one_message(self):
        stream = ReceiveStream(1)
        # incompressed single byte message, followed by two more bytes of /xff
        stream.write(b"\x00\x00\x00\x00\x01\xff\xff\xff")
        # single byte message is queued
        assert stream.queue.get() == (False, b"\xff")
        # following two bytes remain in the buffer
        assert stream.buffer.peek() == b"\xff\xff"

    def test_write_multiple_messages(self):
        stream = ReceiveStream(1)
        for _ in range(10):
            stream.write(b"\x00\x00\x00\x00\x01\xff")  # 10 single byte messages

        assert stream.queue.qsize() == 10
        assert len(stream.buffer) == 0

    def test_consume_grpc_error(self):
        stream = ReceiveStream(1)
        error = GrpcError("boom", "details")
        stream.queue.put(error)

        message_type = Mock()

        with pytest.raises(GrpcError):
            next(stream.consume(message_type))

    def test_consume_end_of_stream(self):
        stream = ReceiveStream(1)
        stream.close()

        message_type = Mock()

        assert list(stream.consume(message_type)) == []

    def test_consume_uncompressed_message(self):
        stream = ReceiveStream(1)

        message_data = b"x"
        message_type = Mock()
        message = message_type()

        stream.queue.put((False, message_data))
        stream.close()  # close stream so that consume exits

        assert list(stream.consume(message_type)) == [message]
        assert message.ParseFromString.call_args_list == [call(message_data)]

    @patch("nameko_grpc.streams.decompress")
    def test_consume_compressed_message(self, decompress):
        stream = ReceiveStream(1)

        message_data = b"x"
        message_type = Mock()
        message = message_type()

        stream.queue.put((True, message_data))
        stream.close()  # close stream so that consume exits

        assert list(stream.consume(message_type)) == [message]
        assert message.ParseFromString.call_args_list == [
            call(decompress(message_data))
        ]

    @patch("nameko_grpc.streams.decompress")
    def test_consume_multiple_messages(self, decompress):
        stream = ReceiveStream(1)

        message_data = b"x"
        message_type = Mock()
        message = message_type()

        stream.queue.put((False, message_data))
        stream.queue.put((True, message_data))
        stream.close()  # close stream so that consume exits

        assert list(stream.consume(message_type)) == [message, message]
        assert message.ParseFromString.call_args_list == [
            call(message_data),
            call(decompress(message_data)),
        ]


class TestSendStream:
    def test_populate(self):
        stream = SendStream(1)
        stream.populate(range(10))

        assert stream.closed
        assert stream.queue.qsize() == 11

    def test_populate_closed_stream(self):
        stream = SendStream(1)
        stream.close()

        assert stream.closed
        stream.populate(range(10))
        assert stream.queue.qsize() == 1


class TestSendStreamHeadersToSend:
    def test_no_headers(self):
        stream = SendStream(1)

        assert len(stream.headers) == 0
        assert stream.headers_to_send(False) is False

    def test_empty_queue(self):
        stream = SendStream(1)
        stream.headers.set(("foo", "bar"))

        assert stream.queue.qsize() == 0
        assert stream.headers_to_send(True) is False
        assert stream.headers_to_send(False) == [(b"foo", b"bar")]

    def test_mark_as_sent(self):
        stream = SendStream(1)
        stream.headers.set(("foo", "bar"))

        assert stream.headers_to_send(False) == [(b"foo", b"bar")]  # marks as sent
        assert stream.headers_to_send(False) is False  # previously sent

    def test_defer_until_data(self):
        stream = SendStream(1)
        stream.headers.set(("foo", "bar"))

        assert stream.headers_to_send(True) is False  # defer until data
        stream.queue.put(Mock())

        assert stream.queue.qsize() == 1
        assert stream.headers_to_send(True) == [(b"foo", b"bar")]


class TestSendStreamTrailersToSend:
    def test_no_trailers(self):
        stream = SendStream(1)

        assert len(stream.trailers) == 0
        assert stream.trailers_to_send() is False

    def test_send_trailers(self):
        stream = SendStream(1)
        stream.trailers.set(("foo", "bar"))

        assert stream.trailers_to_send() == [(b"foo", b"bar")]


@pytest.fixture
def generate_messages():
    with patch("nameko_grpc.streams.compress") as compress:
        compress.side_effect = lambda body, _: (False, body)

        def generate(count, length):
            """ Generate a series of mock messages.

            If `count` is 2 and `length` is 4, when passed to `stream.populate`,
            two messages with the following payload will be added to the stream's
            queue.

                #1. b`\x00\x00\x00\x00`
                #2. b`\x01\x01\x01\x01`
            """
            messages = []
            for index in range(count):
                message = Mock()
                message.SerializeToString.return_value = bytes([index] * length)
                messages.append(message)
            return messages

        yield generate


class TestSendStreamFlushQueueToBuffer:
    def test_empty_queue(self):
        stream = SendStream(1)
        assert stream.queue.qsize() == 0

        stream.flush_queue_to_buffer()
        assert stream.buffer.empty()

    def test_messages_on_queue(self, generate_messages):
        stream = SendStream(1)
        stream.populate(generate_messages(count=2, length=20))

        header = struct.pack("?", False) + struct.pack(">I", 20)

        stream.flush_queue_to_buffer()
        assert stream.buffer.peek() == header + b"\x00" * 20 + header + b"\x01" * 20

    def test_stream_closed(self, generate_messages):
        stream = SendStream(1)
        stream.populate(generate_messages(count=2, length=20))

        header = struct.pack("?", False) + struct.pack(">I", 20)

        stream.flush_queue_to_buffer()
        assert stream.buffer.peek() == header + b"\x00" * 20 + header + b"\x01" * 20

        stream.flush_queue_to_buffer()  # stream closed; no-op
        assert stream.buffer.peek() == header + b"\x00" * 20 + header + b"\x01" * 20

    def test_error_on_queue(self, generate_messages):
        stream = SendStream(1)

        error = GrpcError("boom", "details")
        messages = itertools.chain(generate_messages(count=2, length=20), [error])

        stream.populate(messages)

        with pytest.raises(GrpcError):
            stream.flush_queue_to_buffer()


class TestSendStreamRead:
    def test_no_data(self):
        stream = SendStream(1)

        max_bytes = 10
        chunk_size = 10

        assert stream.buffer.empty()
        assert list(stream.read(max_bytes, chunk_size)) == []

    def test_less_than_one_chunk_of_data(self):
        stream = SendStream(1)
        stream.buffer.write(b"abc")

        max_bytes = 10
        chunk_size = 5

        assert list(stream.read(max_bytes, chunk_size)) == [b"abc"]
        assert stream.buffer.empty()

    def test_more_than_one_chunk_of_data(self):
        stream = SendStream(1)
        stream.buffer.write(b"abcdefghijklm")

        max_bytes = 10
        chunk_size = 5

        assert list(stream.read(max_bytes, chunk_size)) == [b"abcde", b"fghij"]
        assert stream.buffer.peek() == b"klm"

    def test_less_than_max_bytes_of_data(self):
        stream = SendStream(1)
        stream.buffer.write(b"abcdefghijklm")

        max_bytes = 20
        chunk_size = 5

        assert list(stream.read(max_bytes, chunk_size)) == [b"abcde", b"fghij", b"klm"]
        assert stream.buffer.empty()

    def test_more_than_max_bytes_of_data(self):
        stream = SendStream(1)
        stream.buffer.write(b"abcdefghijklm")

        max_bytes = 10
        chunk_size = 5

        assert list(stream.read(max_bytes, chunk_size)) == [b"abcde", b"fghij"]
        assert stream.buffer.peek() == b"klm"

    def test_chunk_greater_than_max_bytes(self):
        stream = SendStream(1)
        stream.buffer.write(b"abcdefghijklm")

        max_bytes = 5
        chunk_size = 10

        assert list(stream.read(max_bytes, chunk_size)) == [b"abcde"]
        assert stream.buffer.peek() == b"fghijklm"

    def test_stream_closed(self):
        stream = SendStream(1)

        max_bytes = 10
        chunk_size = 5

        stream.close()
        assert list(stream.read(max_bytes, chunk_size)) == []

    def test_stream_closed_with_error(self):
        stream = SendStream(1)

        error = GrpcError("boom", "details")
        stream.close(error)

        max_bytes = 10
        chunk_size = 5

        with pytest.raises(GrpcError):
            next(stream.read(max_bytes, chunk_size))

    def test_multiple_small_messages(self, generate_messages):
        stream = SendStream(1)
        stream.populate(generate_messages(count=100, length=1))

        header = struct.pack("?", False) + struct.pack(">I", 1)
        max_bytes = 20
        chunk_size = 10

        chunks = list(stream.read(max_bytes, chunk_size))
        expected = [
            # 5 bytes header + 1 byte payload + 4 bytes of next header
            header + b"\x00" + header[:4],
            # remaining 1 byte of header + 1 byte payload
            # + 5 bytes header + 1 byte payload + 2 bytes of next header
            header[4:] + b"\x01" + header + b"\x02" + header[:2],
        ]
        assert chunks == expected

        assert sum(map(len, chunks)) == max_bytes

        # queue is emptied into buffer
        assert len(stream.buffer) == 100 * (5 + 1) - max_bytes  # 580 bytes left
        assert stream.queue.qsize() == 0

    def test_multiple_large_messages(self, generate_messages):
        stream = SendStream(1)
        stream.populate(generate_messages(count=100, length=200))

        header = struct.pack("?", False) + struct.pack(">I", 200)
        max_bytes = 50
        chunk_size = 10

        chunks = list(stream.read(max_bytes, chunk_size))
        expected = [
            header + b"\x00\x00\x00\x00\x00",  # 5 bytes header + 5 bytes payload
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",  # 10 bytes payload
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",  # 10 bytes payload
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",  # 10 bytes payload
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",  # 10 bytes payload
        ]
        assert chunks == expected

        assert sum(map(len, chunks)) == max_bytes

        # queue is emptied into buffer
        assert len(stream.buffer) == 100 * (5 + 200) - max_bytes  # 20450 bytes left
        assert stream.queue.qsize() == 0

    def test_data_in_buffer_and_messages_in_queue(self, generate_messages):
        stream = SendStream(1)
        stream.buffer.write(b"\xff\xff\xff\xff\xff")
        stream.populate(generate_messages(count=10, length=10))

        header = struct.pack("?", False) + struct.pack(">I", 10)
        max_bytes = 10
        chunk_size = 10

        chunks = list(stream.read(max_bytes, chunk_size))
        expected = [b"\xff\xff\xff\xff\xff" + header]
        assert chunks == expected

        assert sum(map(len, chunks)) == max_bytes

        # queue is emptied into buffer
        assert len(stream.buffer) == 5 + 10 * (5 + 10) - max_bytes  # 145 bytes left
        assert stream.queue.qsize() == 0



# -*- coding: utf-8 -*-
import queue
import socket
from logging import getLogger
from urllib.parse import urlparse

import eventlet

from nameko_grpc.connection import ClientConnectionManager, ServerConnectionManager


log = getLogger(__name__)

CONNECT_TIMEOUT = 5


class ClientConnectionPool:
    """ Simple connection pool for clients.

    Accepts a list of targets and will maintain a connection to each of them,
    round-robining requests between them.

    Currently expects each target to be a valid argument to `urllib.parse.urlparse`.
    If the ClientChannel becomes more complex to support pluggable resolvers and
    load-balancing, `targets` will need more structure. Something like:

        target:
           ip_address: ...
           port: ...
           hostname: ... (for ssl verification)
           service config: ... (maybe)

    """

    def __init__(self, targets, ssl, spawn_thread):
        self.targets = targets
        self.ssl = ssl
        self.spawn_thread = spawn_thread

        self.connections = queue.Queue()

    def connect(self, target):
        sock = socket.create_connection(
            (target.hostname, target.port or 50051), timeout=CONNECT_TIMEOUT
        )

        if self.ssl:
            context = self.ssl.client_context()
            sock = context.wrap_socket(
                sock=sock, server_hostname=target.hostname, suppress_ragged_eofs=True
            )

        sock.settimeout(60)  # XXX needed and/or correct value?
        connection = ClientConnectionManager(sock)
        self.connections.put(connection)

        def run_with_reconnect():
            connection.run_forever()
            if self.run:
                self.connect(target)

        self.spawn_thread(
            target=run_with_reconnect, name=f"grpc client connection [{target}]"
        )

    def get(self):
        while True:
            conn = self.connections.get()
            if conn.alive:
                self.connections.put(conn)
                return conn

    def start(self):
        self.run = True
        for target in self.targets:
            self.connect(urlparse(target))

    def stop(self):
        self.run = False
        while not self.connections.empty():
            self.connections.get().stop()


class ClientChannel:
    """ Simple client channel.

    Channels could eventually suppport pluggable resolvers and load-balancing.
    """

    def __init__(self, target, ssl, spawn_thread):
        self.conn_pool = ClientConnectionPool([target], ssl, spawn_thread)

    def start(self):
        self.conn_pool.start()

    def stop(self):
        self.conn_pool.stop()

    def send_request(self, request_headers):
        return self.conn_pool.get().send_request(request_headers)


class ServerConnectionPool:
    """ Simple connection pool for servers.

    Just accepts new connections and allows them to run until close.
    """

    def __init__(self, host, port, ssl, spawn_thread, handle_request):
        self.host = host
        self.port = port
        self.ssl = ssl
        self.spawn_thread = spawn_thread
        self.handle_request = handle_request

        self.connections = queue.Queue()

    def listen(self):
        sock = eventlet.listen((self.host, self.port))
        sock.settimeout(None)

        if self.ssl:
            context = self.ssl.server_context()
            sock = context.wrap_socket(
                sock=sock, server_side=True, suppress_ragged_eofs=True,
            )

        return sock

    def run(self):
        while self.is_accepting:
            sock, _ = self.listening_socket.accept()
            sock.settimeout(60)  # XXX needed and/or correct value?

            connection = ServerConnectionManager(sock, self.handle_request)
            self.connections.put(connection)
            self.spawn_thread(
                target=connection.run_forever, name=f"grpc server connection [{sock}]"
            )

    def start(self):
        self.listening_socket = self.listen()
        self.is_accepting = True
        self.spawn_thread(
            target=self.run, name=f"grpc server accept [{self.listening_socket}]"
        )

    def stop(self):
        self.is_accepting = False
        while not self.connections.empty():
            self.connections.get().stop()
        self.listening_socket.close()


class ServerChannel:
    """ Simple server channel encapsulating incoming connection management.
    """

    def __init__(self, host, port, ssl, spawn_thread, handle_request):
        self.conn_pool = ServerConnectionPool(
            host, port, ssl, spawn_thread, handle_request
        )

    def start(self):
        self.conn_pool.start()

    def stop(self):
        self.conn_pool.stop()



# -*- coding: utf-8 -*-
import sys
import time
import types
from functools import partial
from logging import getLogger

from grpc import StatusCode
from nameko import config
from nameko.exceptions import ContainerBeingKilled
from nameko.extensions import Entrypoint, SharedExtension, register_entrypoint

from nameko_grpc.channel import ServerChannel
from nameko_grpc.compression import SUPPORTED_ENCODINGS
from nameko_grpc.constants import Cardinality
from nameko_grpc.context import GrpcContext, context_data_from_metadata
from nameko_grpc.errors import GrpcError
from nameko_grpc.inspection import Inspector
from nameko_grpc.ssl import SslConfig
from nameko_grpc.timeout import unbucket_timeout


log = getLogger(__name__)


class GrpcServer(SharedExtension):
    def __init__(self):
        super(GrpcServer, self).__init__()
        self.entrypoints = {}

    def register(self, entrypoint):
        self.entrypoints[entrypoint.method_path] = entrypoint

    def unregister(self, entrypoint):
        self.entrypoints.pop(entrypoint.method_path, None)

    def timeout(self, request_stream, response_stream, deadline):
        start = time.time()
        while True:
            if request_stream.closed and response_stream.closed:
                break
            elapsed = time.time() - start
            if elapsed > deadline:
                request_stream.close()
                error = GrpcError(
                    code=StatusCode.DEADLINE_EXCEEDED, message="Deadline Exceeded"
                )
                response_stream.close(error)
                break
            time.sleep(0.001)

    def handle_request(self, request_stream, response_stream):
        try:
            method_path = request_stream.headers.get(":path")
            entrypoint = self.entrypoints[method_path]
        except KeyError:
            raise GrpcError(code=StatusCode.UNIMPLEMENTED, message="Method not found!")

        encoding = request_stream.headers.get("grpc-encoding", "identity")
        if encoding not in SUPPORTED_ENCODINGS:
            raise GrpcError(
                code=StatusCode.UNIMPLEMENTED,
                message="Algorithm not supported: {}".format(encoding),
            )

        timeout = request_stream.headers.get("grpc-timeout")
        if timeout:
            timeout = unbucket_timeout(timeout)
            self.container.spawn_managed_thread(
                partial(self.timeout, request_stream, response_stream, timeout)
            )

        self.container.spawn_managed_thread(
            partial(entrypoint.handle_request, request_stream, response_stream)
        )

    def setup(self):
        host = config.get("GRPC_BIND_HOST", "0.0.0.0")
        port = config.get("GRPC_BIND_PORT", 50051)
        ssl = SslConfig(config.get("GRPC_SSL"))

        def spawn_thread(target, args=(), kwargs=None, name=None):
            self.container.spawn_managed_thread(
                lambda: target(*args, **kwargs or {}), identifier=name
            )

        self.channel = ServerChannel(host, port, ssl, spawn_thread, self.handle_request)

    def start(self):
        self.channel.start()

    def stop(self):
        self.channel.stop()
        super(GrpcServer, self).stop()

    def kill(self):
        self.stop()


class Grpc(Entrypoint):

    grpc_server = GrpcServer()

    def __init__(self, stub, **kwargs):
        super().__init__(**kwargs)
        self.stub = stub

    @property
    def method_path(self):
        if self.is_bound():
            return Inspector(self.stub).path_for_method(self.method_name)

    @property
    def input_type(self):
        if self.is_bound():
            return Inspector(self.stub).input_type_for_method(self.method_name)

    @property
    def output_type(self):
        if self.is_bound():
            return Inspector(self.stub).output_type_for_method(self.method_name)

    @property
    def cardinality(self):
        if self.is_bound():
            return Inspector(self.stub).cardinality_for_method(self.method_name)

    @classmethod
    def implementing(cls, stub):
        return partial(cls.decorator, stub)

    @classmethod
    def decorator(cls, stub, *args, **kwargs):
        """ Override Entrypoint.decorator to ensure `stub` is passed to instance.

        Would be nicer if Nameko had a better mechanism for this.
        """

        def registering_decorator(fn, args, kwargs):
            instance = cls(stub, *args, **kwargs)
            register_entrypoint(fn, instance)
            return fn

        if len(args) == 1 and isinstance(args[0], types.FunctionType):
            # usage without arguments to the decorator:
            # @foobar
            # def spam():
            #     pass
            return registering_decorator(args[0], args=(), kwargs={})
        else:
            # usage with arguments to the decorator:
            # @foobar('shrub', ...)
            # def spam():
            #     pass
            return partial(registering_decorator, args=args, kwargs=kwargs)

    def setup(self):
        self.grpc_server.register(self)

    def stop(self):
        self.grpc_server.unregister(self)

    def handle_request(self, request_stream, response_stream):

        request = request_stream.consume(self.input_type)

        if self.cardinality in (Cardinality.UNARY_STREAM, Cardinality.UNARY_UNARY):
            request = next(request)

        context = GrpcContext(request_stream, response_stream)

        args = (request, context)
        kwargs = {}

        context_data = context_data_from_metadata(context.invocation_metadata())

        handle_result = partial(self.handle_result, response_stream)
        try:
            self.container.spawn_worker(
                self,
                args,
                kwargs,
                context_data=context_data,
                handle_result=handle_result,
            )
        except ContainerBeingKilled:
            raise GrpcError(
                status=StatusCode.UNAVAILABLE, message="Server shutting down"
            )

    def handle_result(self, response_stream, worker_ctx, result, exc_info):

        if self.cardinality in (Cardinality.STREAM_UNARY, Cardinality.UNARY_UNARY):
            result = (result,)

        if exc_info is None:
            try:
                response_stream.populate(result)
            except Exception as exception:
                message = "Exception iterating responses: {}".format(exception)
                error = GrpcError.from_exception(sys.exc_info(), message=message)

                response_stream.close(error)
        else:
            message = "Exception calling application: {}".format(exc_info[1])
            error = GrpcError.from_exception(exc_info, message=message)

            response_stream.close(error)

        return result, exc_info



# -*- coding: utf-8 -*-
import traceback

from grpc import StatusCode
from nameko import config

from google.protobuf.any_pb2 import Any
from google.rpc.error_details_pb2 import DebugInfo
from google.rpc.status_pb2 import Status


STATUS_CODE_INT_TO_ENUM_MAP = {item.value[0]: item for item in StatusCode}
STATUS_CODE_ENUM_TO_INT_MAP = {item: item.value[0] for item in StatusCode}

GRPC_DETAILS_METADATA_KEY = "grpc-status-details-bin"


class GrpcError(Exception):
    def __init__(self, code, message, status=None):
        self.code = code
        self.message = message
        self.status = status

    def as_headers(self):
        """ Dehydrate this instance to headers to be sent as trailing metadata.
        """
        headers = {
            # ("content-length", "0"),
            "grpc-status": str(STATUS_CODE_ENUM_TO_INT_MAP[self.code]),
            "grpc-message": self.message,
        }
        if self.status:
            headers[GRPC_DETAILS_METADATA_KEY] = self.status.SerializeToString()
        return list(headers.items())

    @classmethod
    def from_headers(cls, headers):
        """ Rehydrate a new instance from headers received as trailing metadata.
        """
        code = int(headers.get("grpc-status"))
        message = headers.get("grpc-message")
        status = headers.get(GRPC_DETAILS_METADATA_KEY)

        return cls(
            code=STATUS_CODE_INT_TO_ENUM_MAP[code],
            message=message,
            status=Status.FromString(status) if status else None,
        )

    @classmethod
    def from_status(cls, status):
        return cls(
            code=STATUS_CODE_INT_TO_ENUM_MAP[status.code],
            message=status.message,
            status=status,
        )

    @staticmethod
    def from_exception(exc_info, code=None, message=None):
        """ Utility method to create a new GrpcError instance representing an
        underlying exception. Useful in try/except clauses.

        By default, a `google.rpc.Status` message will be generated capturing the
        debug info of the underyling traceback. See `default_error_from_exception`.

        This can be overridden by registering a new callable against a given exception
        type. See `register_exception_handler`.
        """
        exc_type, exc, tb = exc_info

        error_from_exception = registry.get(exc_type, default_error_from_exception)
        return error_from_exception(exc_info, code, message)

    def __str__(self):
        return (
            "<RPC terminated with:\n"
            "\tcode = {}\n"
            '\tmessage = "{}"\n'
            '\tstatus = "{}"\n'
            ">".format(self.code, self.message, self.status)
        )


def make_status(code, message, details=None):
    return Status(code=code.value[0], message=message, details=details or [])


def register_exception_handler(exc_type, custom_error_from_exception):
    """ Register a custom implementation to generate a GrpcError from an underlying
    exception, by exception type.

    Must be a callable with a signature matching `default_error_from_exception`.
    """
    registry[exc_type] = custom_error_from_exception


def unregister_expection_handler(exc_type):
    """ Unregister a custom implementation.
    """
    registry.pop(exc_type, None)


def default_error_from_exception(exc_info, code=None, message=None):
    """ Create a new GrpcError instance representing an underlying exception.

    If the `GRPC_DEBUG` key is set in the Nameko config, the `status` message will
    capture the underyling traceback in a `google.rpc.error_details.DebugInfo` message.
    """
    exc_type, exc, tb = exc_info

    code = code or StatusCode.UNKNOWN
    message = message or str(exc)

    status = Status(code=STATUS_CODE_ENUM_TO_INT_MAP[code], message=message)

    if config.get("GRPC_DEBUG"):
        debug_info = Any()
        debug_info.Pack(
            DebugInfo(
                stack_entries=traceback.format_exception(*exc_info), detail=str(exc),
            )
        )
        status.details.append(debug_info)

    return GrpcError(code=code, message=message, status=status)


def grpc_error_passthrough_exception_handler(exc_info, code, message):
    exc_type, exc, tb = exc_info
    return exc


registry = {GrpcError: grpc_error_passthrough_exception_handler}



# -*- coding: utf-8 -*-
import pytest

from nameko_grpc.constants import Cardinality
from nameko_grpc.inspection import Inspector


class TestInspection:
    @pytest.fixture
    def inspector(self, stubs):
        return Inspector(stubs.exampleStub)

    def test_service_name(self, inspector):
        assert inspector.service_name == "nameko.example"

    def test_path_for_method(self, inspector):
        assert inspector.path_for_method("unary_unary") == "/nameko.example/unary_unary"
        assert (
            inspector.path_for_method("unary_stream") == "/nameko.example/unary_stream"
        )
        assert (
            inspector.path_for_method("stream_stream")
            == "/nameko.example/stream_stream"
        )
        assert (
            inspector.path_for_method("stream_unary") == "/nameko.example/stream_unary"
        )

    def test_input_type_for_method(self, inspector, protobufs):
        assert (
            inspector.input_type_for_method("unary_unary") == protobufs.ExampleRequest
        )

    def test_output_type_for_method(self, inspector, protobufs):
        assert inspector.output_type_for_method("unary_unary") == protobufs.ExampleReply

    def test_cardinality_for_method(self, inspector):
        insp = inspector
        assert insp.cardinality_for_method("unary_unary") == Cardinality.UNARY_UNARY
        assert insp.cardinality_for_method("unary_stream") == Cardinality.UNARY_STREAM
        assert insp.cardinality_for_method("stream_unary") == Cardinality.STREAM_UNARY
        assert insp.cardinality_for_method("stream_stream") == Cardinality.STREAM_STREAM



# -*- coding: utf-8 -*-
import base64
import json
import os
import pickle
import threading
import time

import grpc
import wrapt
import zmq
from google.rpc import status_pb2

from nameko_grpc.constants import Cardinality
from nameko_grpc.errors import GRPC_DETAILS_METADATA_KEY, GrpcError


def status_from_metadata(metadata):
    for key, value in metadata:
        if key == GRPC_DETAILS_METADATA_KEY:
            return status_pb2.Status.FromString(value)


def extract_metadata(context):
    """ Extracts invocation metadata from `context` and returns it as JSON.

    Binary headers are included as base64 encoded strings.
    Duplicate header values are joined as comma separated, preserving order.
    """
    metadata = {}
    for key, value in context.invocation_metadata():
        if key.endswith("-bin"):
            value = base64.b64encode(value).decode("utf-8")
        if key in metadata:
            metadata[key] = "{},{}".format(metadata[key], value)
        else:
            metadata[key] = value
    return json.dumps(metadata)


def maybe_echo_metadata(context):
    """ Copy metadata from request to response.
    """
    initial_metadata = []
    trailing_metadata = []

    for key, value in context.invocation_metadata():
        if key.startswith("echo-header"):
            initial_metadata.append((key[12:], value))
        elif key.startswith("echo-trailer"):
            trailing_metadata.append((key[13:], value))

    if initial_metadata:
        context.send_initial_metadata(initial_metadata)
    if trailing_metadata:
        context.set_trailing_metadata(trailing_metadata)


def maybe_sleep(request):
    if request.delay:
        time.sleep(request.delay / 1000)


class RemoteClientTransport:
    """ Serialializes/deserializes objects through ZMQ sockets using `pickle`.
    """

    ENDSTREAM = "endstream"

    def __init__(self, sock):
        self.sock = sock

    @classmethod
    def bind(cls, context, socket_type, target):
        """ Create a new transport over a ZMQ socket of type `socket_type`, bound
        to the `target` address.
        """
        socket = context.socket(socket_type)
        socket.bind(target)
        return RemoteClientTransport(socket)

    @classmethod
    def bind_to_free_port(cls, context, socket_type, target):
        socket = context.socket(socket_type)
        port = socket.bind_to_random_port(target)
        return RemoteClientTransport(socket), port

    @classmethod
    def connect(cls, context, socket_type, target):
        """ Create a new transport over a ZMQ socket of type `socket_type`, connected
        to the `target` address.
        """
        socket = context.socket(socket_type)
        socket.connect(target)
        return RemoteClientTransport(socket)

    @property
    def context(self):
        return self.sock.context

    def receive(self, close=False):
        loaded = pickle.loads(self.sock.recv())
        try:
            if isinstance(loaded, GrpcError):
                raise loaded
            return loaded
        finally:
            if close:
                self.close()

    def receive_stream(self):
        while True:
            item = self.receive()
            if item == self.ENDSTREAM:
                break
            yield item
        self.close()

    def send(self, result, close=False):
        self.sock.send(pickle.dumps(result))
        if close:
            self.close()

    def send_stream(self, result):
        try:
            for item in result:
                self.send(item)
        except grpc.RpcError as exc:
            state = exc._state
            error = GrpcError(
                state.code, state.details, status_from_metadata(state.trailing_metadata)
            )
            self.send(error)
        self.send(self.ENDSTREAM, close=True)

    def close(self):
        self.sock.close()


class Command:
    """ Encapsulates a command to run on a remote GRPC client.

    A command can be issued to a remote client, or responded to by that remote client.

    Issuing a command involves serializing it with pickle and sending it to the
    remote client via ZeroMQ (serialization is by a `RemoteClientTransport`)

    At the remote client, the received `Command` can be used to receive the request to
    issue to the GRPC server, and to return the response from the server to the caller.

    Commands have two modes: ISSUE and RESPOND. A Command is constructed in ISSUE mode,
    and converted to RESPOND mode only when it is deserialised by pickle at the remote
    client.

    When in ISSUE mode, the Command can send requests to and receive responses from the
    remote client. When in RESPOND mode, the Command can receive requests from and send
    responses back to the caller.
    """

    END = "endcommands"

    class Modes:
        ISSUE = "issue"
        RESPOND = "respond"

    def __init__(self, method_name, cardinality, kwargs, transport):
        self.method_name = method_name
        self.cardinality = cardinality
        self.kwargs = kwargs
        self.transport = transport

        self.request_port = None
        self.response_port = None
        self.metadata_port = None

        self._request_transport = None
        self._response_transport = None
        self._metadata_transport = None

        self.mode = Command.Modes.ISSUE

    def __getstate__(self):
        # remove the local transports before pickling
        state = self.__dict__.copy()
        state["transport"] = None
        state["_request_transport"] = None
        state["_response_transport"] = None
        state["_metadata_transport"] = None
        return state

    def __setstate__(self, newstate):
        # set the mode to RESPOND when unpickling
        newstate["mode"] = Command.Modes.RESPOND
        self.__dict__.update(newstate)

    def issue(self):
        """ Issue this command to a remote client.

        Sends itelf over the given `transport`.
        """
        if self.mode is not Command.Modes.ISSUE:
            raise ValueError("Command must be in ISSUE mode to be issued")

        # ensure transports are established and bound to ports before transmitting
        assert self.request_transport
        assert self.response_transport
        assert self.metadata_transport

        self.transport.send(self)
        assert self.transport.receive() is True

    @staticmethod
    def retrieve_commands(transport):
        """ Retrieve a series of commands over the given `transport`
        """
        # this is actually just receive_stream...
        while True:
            command = transport.receive()
            if command == Command.END:
                break
            assert isinstance(command, Command)
            command.transport = transport
            yield command
            transport.send(True)
        transport.close()

    def bind_transport(self, socket_type):
        """ Bind a new transport using the given `socket_type`.
        """
        transport, port = RemoteClientTransport.bind_to_free_port(
            self.transport.context, socket_type, "tcp://127.0.0.1"
        )
        return transport, port

    def connect_transport(self, socket_type, port):
        """ Connect a new transport to `port` using the given `socket_type`.
        """
        transport = RemoteClientTransport.connect(
            self.transport.context, socket_type, "tcp://127.0.0.1:{}".format(port)
        )
        return transport

    @property
    def request_transport(self):
        """ Bind or connect the request transport.

        Depending on the mode of this command, either binds a new transport to a free
        port, or connects to the established port.
        """
        if self._request_transport is None:
            if self.mode is Command.Modes.ISSUE:
                transport, self.request_port = self.bind_transport(zmq.PUSH)
            else:
                transport = self.connect_transport(zmq.PULL, self.request_port)
            self._request_transport = transport
        return self._request_transport

    @property
    def response_transport(self):
        """ Bind or connect the response transport.

        Depending on the mode of this command, either binds a new transport to a free
        port, or connects to the established port.
        """
        if self._response_transport is None:
            if self.mode is Command.Modes.ISSUE:
                transport, self.response_port = self.bind_transport(zmq.PULL)
            else:
                transport = self.connect_transport(zmq.PUSH, self.response_port)
            self._response_transport = transport
        return self._response_transport

    @property
    def metadata_transport(self):
        """ Bind or connect the metadata transport.

        Depending on the mode of this command, either binds a new transport to a free
        port, or connects to the established port.
        """
        if self._metadata_transport is None:
            if self.mode is Command.Modes.ISSUE:
                transport, self.metadata_port = self.bind_transport(zmq.PULL)
            else:
                transport = self.connect_transport(zmq.PUSH, self.metadata_port)
            self._metadata_transport = transport
        return self._metadata_transport

    def send_request(self, request):
        """ Send the request for this command to the remote client.

        Only available in ISSUE mode.
        """
        if self.mode is not Command.Modes.ISSUE:
            raise ValueError("Command must be in ISSUE mode to send request")

        if self.cardinality in (Cardinality.STREAM_UNARY, Cardinality.STREAM_STREAM):
            threading.Thread(
                target=self.request_transport.send_stream, args=(request,)
            ).start()
        else:
            self.request_transport.send(request, close=True)

    def get_request(self):
        """ Get the request for this command from the caller.

        Only available in RESPOND mode.
        """
        if self.mode is not Command.Modes.RESPOND:
            raise ValueError("Command must be in RESPOND mode to get request")

        if self.cardinality in (Cardinality.STREAM_UNARY, Cardinality.STREAM_STREAM):
            return self.request_transport.receive_stream()
        return self.request_transport.receive(close=True)

    def send_response(self, response):
        """ Send the response for this command
        """
        if self.mode is not Command.Modes.RESPOND:
            raise ValueError("Command must be in RESPOND mode to send a response")

        if self.cardinality in (Cardinality.UNARY_STREAM, Cardinality.STREAM_STREAM):
            threading.Thread(
                target=self.response_transport.send_stream, args=(response,)
            ).start()
        else:
            self.response_transport.send(response, close=True)

    def get_response(self):
        """ Get the response for this command from the remote client
        """
        if self.mode is not Command.Modes.ISSUE:
            raise ValueError("Command must be in ISSUE mode to get a response")

        if self.cardinality in (Cardinality.UNARY_STREAM, Cardinality.STREAM_STREAM):
            return self.response_transport.receive_stream()
        return self.response_transport.receive(close=True)

    def send_metadata(self, metadata):
        """ Send the response metadata for this command
        """
        if self.mode is not Command.Modes.RESPOND:
            raise ValueError("Command must be in RESPOND mode to send metadata")

        self.metadata_transport.send(metadata, close=True)

    def get_metadata(self):
        """ Get the response metadata for this command from the remote client
        """
        if self.mode is not Command.Modes.ISSUE:
            raise ValueError("Command must be in ISSUE mode to get a metadata")

        return self.metadata_transport.receive(close=True)


class Stash:

    REQUEST = "REQ"
    RESPONSE = "RES"

    def __init__(self, path):
        self.path = path

    def write(self, value, type_):
        if not self.path:
            return

        with open(self.path, "ab") as fh:
            fh.write(pickle.dumps((type_, value)) + b"|")

    def stash_request(self, request):
        self.write(request, self.REQUEST)
        return request

    def stash_request_stream(self, iterator):
        for request in iterator:
            self.write(request, self.REQUEST)
            yield request

    def stash_response(self, response):
        self.write(response, self.RESPONSE)
        return response

    def stash_response_stream(self, iterator):
        for response in iterator:
            self.write(response, self.RESPONSE)
            yield response

    def read(self):
        if not self.path or not os.path.exists(self.path):
            return []

        with open(self.path, "rb") as fh:
            data = fh.read()

        for item in data.split(b"|"):
            if item:
                yield pickle.loads(item)

    def requests(self):
        for type_, item in self.read():
            if type_ == self.REQUEST:
                yield item

    def responses(self):
        for type_, item in self.read():
            if type_ == self.RESPONSE:
                yield item


@wrapt.decorator
def instrumented(wrapped, instance, args, kwargs):
    """ Decorator for instrumenting GRPC implementation methods.

    Stores requests and responses to file for later inspection.
    """
    request, context = args

    stash_metadata = dict(context.invocation_metadata()).get("stash")
    if stash_metadata:
        stash_path, cardinality = json.loads(stash_metadata)
    else:
        stash_path, cardinality = None, None

    stash = Stash(stash_path)

    if cardinality in (Cardinality.STREAM_UNARY.value, Cardinality.STREAM_STREAM.value):
        request = stash.stash_request_stream(request)
    else:
        request = stash.stash_request(request)

    response = wrapped(request, context, **kwargs)

    if cardinality in (Cardinality.UNARY_STREAM.value, Cardinality.STREAM_STREAM.value):
        response = stash.stash_response_stream(response)
    else:
        response = stash.stash_response(response)

    return response



# -*- coding: utf-8 -*-
import importlib
import inspect

from google.protobuf import descriptor
from mock import Mock

from nameko_grpc.constants import Cardinality


class Inspector:
    _stub_module = None
    _protobufs_module = None
    _service_descriptor = None
    _method_descriptors = None
    _cardinality_map = None

    def __init__(self, stub):
        self.stub = stub

    @property
    def stub_module(self):
        if self._stub_module is None:
            self._stub_module = importlib.import_module(self.stub.__module__)
        return self._stub_module

    @property
    def protobufs_module(self):
        if self._protobufs_module is None:
            self._protobufs_module = importlib.import_module(self.stub.__module__[:-5])
        return self._protobufs_module

    @property
    def service_descriptor(self):
        if self._service_descriptor is None:
            self._service_descriptor = inspect.getmembers(
                self.protobufs_module,
                lambda member: isinstance(member, descriptor.ServiceDescriptor)
                and member.name == self.stub.__name__[:-4],
            )[0][1]
        return self._service_descriptor

    @property
    def method_descriptors(self):
        if self._method_descriptors is None:
            self._method_descriptors = {
                name: descriptor
                for name, descriptor in self.service_descriptor.methods_by_name.items()
            }
        return self._method_descriptors

    @property
    def cardinality_map(self):
        if self._cardinality_map is None:
            cmap = {}

            mock_channel = Mock()
            self.stub(mock_channel)

            for (method_path,), _ in mock_channel.unary_unary.call_args_list:
                cmap[method_path.split("/")[-1]] = Cardinality.UNARY_UNARY

            for (method_path,), _ in mock_channel.unary_stream.call_args_list:
                cmap[method_path.split("/")[-1]] = Cardinality.UNARY_STREAM

            for (method_path,), _ in mock_channel.stream_unary.call_args_list:
                cmap[method_path.split("/")[-1]] = Cardinality.STREAM_UNARY

            for (method_path,), _ in mock_channel.stream_stream.call_args_list:
                cmap[method_path.split("/")[-1]] = Cardinality.STREAM_STREAM

            self._cardinality_map = cmap

        return self._cardinality_map

    @property
    def service_name(self):
        return self.service_descriptor.full_name

    def get_symbol(self, name):
        return self.protobufs_module._sym_db.GetSymbol(name)

    def path_for_method(self, method_name):
        return "/{}/{}".format(self.service_name, method_name)

    def input_type_for_method(self, method_name):
        return self.get_symbol(
            self.method_descriptors[method_name].input_type.full_name
        )

    def output_type_for_method(self, method_name):
        return self.get_symbol(
            self.method_descriptors[method_name].output_type.full_name
        )

    def cardinality_for_method(self, method_name):
        return self.cardinality_map[method_name]



# -*- coding: utf-8 -*-
import base64

import pytest

from nameko_grpc.headers import (
    HeaderManager,
    check_decoded,
    check_encoded,
    comma_join,
    decode_header,
    encode_header,
    filter_headers_for_application,
    sort_headers_for_wire,
)


class TestEncodeHeader:
    def test_binary(self):
        assert encode_header(("foo-bin", b"123")) == (
            b"foo-bin",
            base64.b64encode(b"123"),
        )

    def test_string_value(self):
        assert encode_header(("foo", "123")) == (b"foo", b"123")


class TestDecodeHeader:
    def test_binary(self):
        assert decode_header((b"foo-bin", base64.b64encode(b"123"))) == (
            "foo-bin",
            b"123",
        )

    def test_binary_with_truncated_padding(self):
        padded_value = base64.b64encode(b"1234")
        assert padded_value.endswith(b"=")

        trimmed_value = padded_value[:-2]
        assert decode_header((b"foo-bin", trimmed_value)) == ("foo-bin", b"1234",)

    def test_string_value(self):
        assert decode_header((b"foo", b"123")) == ("foo", "123")


class TestFilterHeadersForApplication:
    def test_no_application(self):
        headers = [(":status", "1"), ("content-type", "2"), ("grpc-foo", "3")]
        assert filter_headers_for_application(headers) == []

    def test_all_application(self):
        headers = [("foo", "1"), ("bar", "2"), ("baz", "3")]
        assert filter_headers_for_application(headers) == headers

    def test_filter(self):
        headers = [
            ("foo", "1"),
            (":status", "1"),
            ("bar", "2"),
            ("content-type", "2"),
            ("baz", "3"),
        ]
        assert filter_headers_for_application(headers) == [
            ("foo", "1"),
            ("bar", "2"),
            ("baz", "3"),
        ]


class TestSortHeadersForWire:
    def test_empty(self):
        unsorted = []
        for_wire = []
        assert sort_headers_for_wire(unsorted) == for_wire

    def test_already_sorted(self):
        unsorted = [
            (":status", "1"),
            ("content-type", "2"),
            ("grpc-foo", "3"),
            ("other", "4"),
        ]
        for_wire = [
            (":status", "1"),
            ("content-type", "2"),
            ("grpc-foo", "3"),
            ("other", "4"),
        ]
        assert sort_headers_for_wire(unsorted) == for_wire

    def test_sort(self):
        unsorted = [
            ("content-type", "2"),
            (":status", "1"),
            ("other", "4"),
            ("grpc-foo", "3"),
        ]
        for_wire = [
            (":status", "1"),
            ("content-type", "2"),
            ("grpc-foo", "3"),
            ("other", "4"),
        ]
        assert sort_headers_for_wire(unsorted) == for_wire

    def test_multi_sort(self):
        unsorted = [
            ("content-type", "1"),
            ("te", "2"),
            (":status", "3"),
            (":authority", "4"),
            ("other", "5"),
            ("grpc-foo", "6"),
            ("grpc-bar", "7"),
            ("more", "8"),
            (":method", "9"),
        ]
        for_wire = [
            (":status", "3"),
            (":authority", "4"),
            (":method", "9"),
            ("content-type", "1"),
            ("te", "2"),
            ("grpc-foo", "6"),
            ("grpc-bar", "7"),
            ("other", "5"),
            ("more", "8"),
        ]
        assert sort_headers_for_wire(unsorted) == for_wire


class TestCheckEncoded:
    def test_empty(self):
        assert check_encoded([]) is None

    def test_good(self):
        assert check_encoded([(b"foo", b"bar")]) is None

    def test_bad_name(self):
        with pytest.raises(AssertionError):
            check_encoded([("foo", b"bar")])

    def test_bad_value(self):
        with pytest.raises(AssertionError):
            check_encoded([(b"foo", "bar")])


class TestCheckDecoded:
    def test_empty(self):
        assert check_decoded([]) is None

    def test_good(self):
        assert check_decoded([("foo", "bar")]) is None

    def test_bad_name(self):
        with pytest.raises(AssertionError):
            check_decoded([(b"foo", "bar")])

    def test_bad_value(self):
        with pytest.raises(AssertionError):
            check_decoded([("foo", b"bar")])

    def test_good_binary(self):
        assert check_decoded([("foo-bin", b"bar")]) is None

    def test_bad_binary(self):
        with pytest.raises(AssertionError):
            check_decoded([("foo-bin", "bar")])


class TestCommaJoin:
    def test_string(self):
        assert comma_join(["foo", "bar"]) == "foo,bar"

    def test_bytes(self):
        assert comma_join([b"foo", b"bar"]) == b"foo,bar"

    def test_mixed(self):
        with pytest.raises(TypeError):
            comma_join([b"foo", "bar"])


class TestHeaderManager:
    def test_get(self):
        manager = HeaderManager()
        manager.set(("foo", "bar"))

        assert manager.get("foo") == "bar"

    def test_get_with_default(self):
        manager = HeaderManager()

        assert manager.get("foo", "baz") == "baz"

    def test_get_multi(self):
        manager = HeaderManager()
        manager.set(("foo", "bar"), ("foo", "baz"))

        assert manager.get("foo") == "bar,baz"

    def test_set(self):
        manager = HeaderManager()
        manager.set(("x", "y"), ("foo", "bar"))

        assert manager.get("x") == "y"
        assert manager.get("foo") == "bar"
        manager.set(("foo", "baz"))  # clears existing foo, only
        assert manager.get("foo") == "baz"
        assert manager.get("x") == "y"

    def test_set_from_wire(self):
        manager = HeaderManager()

        manager.set((b"foo", b"bar"), from_wire=True)
        assert manager.get("foo") == "bar"

    def test_append(self):
        manager = HeaderManager()
        manager.set(("x", "y"), ("foo", "bar"))

        assert manager.get("x") == "y"
        assert manager.get("foo") == "bar"
        manager.append(("foo", "baz"))  # appends to foo
        assert manager.get("foo") == "bar,baz"
        assert manager.get("x") == "y"

    def test_append_from_wire(self):
        manager = HeaderManager()
        manager.set(("foo", "bar"))

        manager.append((b"foo", b"baz"), from_wire=True)
        assert manager.get("foo") == "bar,baz"

    def test_for_wire(self):
        manager = HeaderManager()
        manager.set(("x", "y"), ("foo", "bar"))

        assert manager.for_wire == [(b"x", b"y"), (b"foo", b"bar")]



# -*- coding: utf-8 -*-
import pytest


@pytest.mark.secure
@pytest.mark.equivalence
class TestStandard:
    def test_unary_unary(self, client, protobufs):
        response = client.unary_unary(protobufs.ExampleRequest(value="A"))
        assert response.message == "A"

    def test_unary_stream(self, client, protobufs):
        responses = client.unary_stream(
            protobufs.ExampleRequest(value="A", response_count=2)
        )
        assert [(response.message, response.seqno) for response in responses] == [
            ("A", 1),
            ("A", 2),
        ]

    def test_stream_unary(self, client, protobufs):
        def generate_requests():
            for value in ["A", "B"]:
                yield protobufs.ExampleRequest(value=value)

        response = client.stream_unary(generate_requests())
        assert response.message == "A,B"

    def test_stream_stream(self, client, protobufs):
        def generate_requests():
            for value in ["A", "B"]:
                yield protobufs.ExampleRequest(value=value)

        responses = client.stream_stream(generate_requests())
        assert [(response.message, response.seqno) for response in responses] == [
            ("A", 1),
            ("B", 2),
        ]



# -*- coding: utf-8 -*-
import sys
import time
from concurrent import futures
from importlib import import_module

import grpc
from grpc._cython.cygrpc import CompressionAlgorithm, CompressionLevel


_ONE_DAY_IN_SECONDS = 60 * 60 * 24


if __name__ == "__main__":

    port = sys.argv[1]

    secure = sys.argv[2]

    proto_name = sys.argv[3]
    service_name = sys.argv[4]

    compression_algorithm = sys.argv[5]
    compression_level = sys.argv[6]

    service_module = import_module("{}_grpc".format(proto_name))
    service_cls = getattr(service_module, service_name)

    grpc_module = import_module("{}_pb2_grpc".format(proto_name))
    add_servicer = getattr(grpc_module, "add_{}Servicer_to_server".format(service_name))

    server_options = [
        (
            "grpc.default_compression_algorithm",
            getattr(CompressionAlgorithm, compression_algorithm),
        ),
        (
            "grpc.default_compression_level",
            getattr(CompressionLevel, compression_level),
        ),
    ]

    def serve():

        server = grpc.server(
            futures.ThreadPoolExecutor(max_workers=10), options=server_options
        )
        add_servicer(service_cls(), server)
        if secure == "secure":

            with open("test/certs/server.key", "rb") as f:
                private_key = f.read()
            with open("test/certs/server.crt", "rb") as f:
                certificate_chain = f.read()

            creds = grpc.ssl_server_credentials(((private_key, certificate_chain),))
            server.add_secure_port("[::]:{}".format(port), creds)
        else:
            server.add_insecure_port("[::]:{}".format(port))

        server.start()
        try:
            while True:
                time.sleep(_ONE_DAY_IN_SECONDS)
        except KeyboardInterrupt:
            server.stop(0)

    serve()



# -*- coding: utf-8 -*-
import ssl


DEFAULT_VERIFY_MODE = "required"
DEFAULT_CHECK_HOSTNAME = True
DEFAULT_SSL_CONFIG = {
    "verify_mode": DEFAULT_VERIFY_MODE,
    "check_hostname": DEFAULT_CHECK_HOSTNAME,
}


class SslConfig:
    def __init__(self, config):
        """ Valid values for `config` are:

            - True
            - False
            - A dict with the following format, all keys optional:

                {
                    "verify_mode": <"none"|"optional"|"required">,
                    "check_hostname": <True|False>,
                    "verify_locations": {
                        <args for SslContext.load_verify_locations>
                    },
                    "cert_chain": {
                        <args for SslContext.load_cert_chain>
                    }
                }
        """
        # TODO add schema for this config dict
        if config is True:
            config = DEFAULT_SSL_CONFIG
        self.config = config

    def __bool__(self):
        return bool(self.config)

    @property
    def verify_mode(self):
        if not self.config:
            return None

        value = self.config.get("verify_mode", DEFAULT_VERIFY_MODE)
        verify_modes = {
            "none": ssl.CERT_NONE,
            "optional": ssl.CERT_OPTIONAL,
            "required": ssl.CERT_REQUIRED,
        }
        return verify_modes[value]

    @property
    def check_hostname(self):
        if not self.config:
            return None

        return self.config.get("check_hostname", DEFAULT_CHECK_HOSTNAME)

    @property
    def verify_locations(self):
        return self.config.get("verify_locations", None)

    @property
    def cert_chain(self):
        return self.config.get("cert_chain", None)

    def server_context(self):
        """ Returns a configured context for use in a server.
        """
        context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        if self.cert_chain:
            context.load_cert_chain(**self.cert_chain)
        context.set_alpn_protocols(["h2"])
        return context

    def client_context(self):
        """ Returns a configured context for use in a client.
        """
        context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        context.check_hostname = self.check_hostname
        context.verify_mode = self.verify_mode
        context.set_alpn_protocols(["h2"])
        return context



# -*- coding: utf-8 -*-
from logging import getLogger

from nameko import config
from nameko.extensions import DependencyProvider

from nameko_grpc.client import ClientBase, Method
from nameko_grpc.context import metadata_from_context_data


log = getLogger(__name__)


class Proxy:
    def __init__(self, client, context_data):
        self.client = client
        self.context_data = context_data

    def __getattr__(self, name):
        extra_metadata = metadata_from_context_data(self.context_data)
        return Method(self.client, name, extra_metadata)


class GrpcProxy(ClientBase, DependencyProvider):
    def __init__(self, *args, **kwargs):
        ssl = kwargs.pop("ssl", config.get("GRPC_SSL"))
        super().__init__(*args, ssl=ssl, **kwargs)

    def spawn_thread(self, target, args=(), kwargs=None, name=None):
        self.container.spawn_managed_thread(
            lambda: target(*args, **kwargs or {}), identifier=name
        )

    def get_dependency(self, worker_ctx):
        return Proxy(self, worker_ctx.context_data)



# -*- coding: utf-8 -*-
import subprocess

import pytest


@pytest.fixture(autouse=True, scope="session")
def interop_protos(compile_proto):
    compile_proto("empty")
    compile_proto("messages")
    compile_proto("interop")


@pytest.fixture
def grpc_server(start_grpc_server):
    return start_grpc_server("TestService", "interop")


@pytest.fixture
def nameko_server(start_nameko_server):
    return start_nameko_server("TestService", "interop")


@pytest.fixture
def server(request, server_type):
    if server_type == "grpc":
        if request.config.option.server not in ("grpc", "all"):
            pytest.skip("grpc server not requested")
        return request.getfixturevalue("grpc_server")
    if server_type == "nameko":
        if request.config.option.server not in ("nameko", "all"):
            pytest.skip("nameko server not requested")
        return request.getfixturevalue("nameko_server")


@pytest.mark.parametrize(
    "testcase,skip_if_grpc,skip_if_nameko",
    [
        ("large_unary", False, False),
        ("unimplemented_service", False, False),
        ("unimplemented_method", False, False),
        ("client_streaming", False, False),
        ("server_streaming", False, False),
        ("server_compressed_unary", True, True),  # compression not supported
        ("server_compressed_streaming", True, True),  # compression not supported
        ("client_compressed_unary", False, False),
        ("client_compressed_streaming", True, True),  # both servers hang, not sure why
        ("empty_unary", False, False),
        ("empty_stream", False, False),
        ("long_lived_channel", True, True),  # takes forever
        ("cacheable_unary", True, True),  # irrelevant, no caching proxy in use
        ("timeout_on_sleeping_server", False, False),
        ("slow_consumer", False, False),
        ("half_duplex", False, False),
        ("ping_pong", False, False),
        ("rpc_soak", False, False),
        ("channel_soak", False, False),
        ("cancel_after_begin", False, False),
        ("cancel_after_first_response", False, False),
        ("custom_metadata", False, False),
        ("status_code_and_message", False, False),
    ],
)
def test_interop(
    server, grpc_port, testcase, skip_if_grpc, skip_if_nameko, server_type
):
    if server_type == "nameko" and skip_if_nameko:
        pytest.skip("Server not compatible")
    if server_type == "grpc" and skip_if_grpc:
        pytest.skip("Server not compatible")

    args = [
        "docker",
        "run",
        "nameko/nameko-grpc-interop",
        "-server_port",
        str(grpc_port),
        "-test_case",
        testcase,
    ]
    assert subprocess.call(args) == 0



# -*- coding: utf-8 -*-
import itertools
import select
import sys
from collections import deque
from contextlib import contextmanager
from logging import getLogger
from threading import Event

from grpc import StatusCode
from h2.config import H2Configuration
from h2.connection import H2Connection
from h2.errors import ErrorCodes
from h2.events import (
    ConnectionTerminated,
    DataReceived,
    RemoteSettingsChanged,
    RequestReceived,
    ResponseReceived,
    SettingsAcknowledged,
    StreamEnded,
    StreamReset,
    TrailersReceived,
    WindowUpdated,
)
from h2.exceptions import StreamClosedError

from nameko_grpc.compression import (
    SUPPORTED_ENCODINGS,
    UnsupportedEncoding,
    select_algorithm,
)
from nameko_grpc.errors import GrpcError
from nameko_grpc.streams import ReceiveStream, SendStream


log = getLogger(__name__)


SELECT_TIMEOUT = 0.01


class ConnectionManager:
    """
    Base class for managing a single GRPC HTTP/2 connection.

    Provides the H2 event loop that manages the connection over a socket. Methods for
    handling each HTTP2 event are fully or partially implemented and can be extended
    by subclasses.
    """

    def __init__(self, sock, client_side):
        self.sock = sock

        config = H2Configuration(client_side=client_side)
        self.conn = H2Connection(config=config)

        self.receive_streams = {}
        self.send_streams = {}

        self.run = True
        self.stopped = Event()

    @property
    def alive(self):
        return not self.stopped.is_set()

    @contextmanager
    def cleanup_on_exit(self):
        error = None
        try:
            yield
        except Exception:
            log.info(
                "ConnectionManager shutting down with error. Traceback:", exc_info=True,
            )
            error = GrpcError.from_exception(
                sys.exc_info(), code=StatusCode.UNAVAILABLE
            )
        finally:
            for send_stream in self.send_streams.values():
                if send_stream.closed:
                    continue  # stream.close() is idemponent but this prevents the log
                log.info(
                    f"Terminating send stream {send_stream}"
                    f"{f' with error {error}' if error else ''}."
                )
                send_stream.close(error)
            for receive_stream in self.receive_streams.values():
                if receive_stream.closed:
                    continue  # stream.close() is idemponent but this prevents the log
                log.info(
                    f"Terminating receive stream {receive_stream}"
                    f"{f' with error {error}' if error else ''}."
                )
                receive_stream.close(error)
            self.stopped.set()

    def run_forever(self):
        """ Event loop.
        """
        self.conn.initiate_connection()

        with self.cleanup_on_exit():

            while self.run:

                self.on_iteration()

                self.sock.sendall(self.conn.data_to_send())
                ready = select.select([self.sock], [], [], SELECT_TIMEOUT)
                if not ready[0]:
                    continue

                data = self.sock.recv(65535)
                if not data:
                    break

                events = self.conn.receive_data(data)

                for event in events:
                    if isinstance(event, RequestReceived):
                        self.request_received(event)
                    elif isinstance(event, ResponseReceived):
                        self.response_received(event)
                    elif isinstance(event, DataReceived):
                        self.data_received(event)
                    elif isinstance(event, StreamEnded):
                        self.stream_ended(event)
                    elif isinstance(event, StreamReset):
                        self.stream_reset(event)
                    elif isinstance(event, WindowUpdated):
                        self.window_updated(event)
                    elif isinstance(event, RemoteSettingsChanged):
                        self.settings_changed(event)
                    elif isinstance(event, SettingsAcknowledged):
                        self.settings_acknowledged(event)
                    elif isinstance(event, TrailersReceived):
                        self.trailers_received(event)
                    elif isinstance(event, ConnectionTerminated):
                        self.connection_terminated(event)

    def stop(self):
        self.run = False
        self.stopped.wait()
        self.sock.close()

    def on_iteration(self):
        """ Called on every iteration of the event loop.

        If there are any open `SendStream`s with headers or data to send, try to send
        them.
        """
        for stream_id in list(self.send_streams.keys()):
            self.send_headers(stream_id)
            self.send_data(stream_id)

    def request_received(self, event):
        """ Called when a request is received on a stream.

        Subclasses should extend this method to handle the request accordingly.
        """
        log.debug("request received, stream %s", event.stream_id)

    def response_received(self, event):
        """ Called when a response is received on a stream.

        Subclasses should extend this method to handle the response accordingly.
        """
        log.debug("response received, stream %s", event.stream_id)
        receive_stream = self.receive_streams.get(event.stream_id)
        if not receive_stream:
            return

        receive_stream.headers.set(*event.headers, from_wire=True)

    def data_received(self, event):
        """ Called when data is received on a stream.

        If there is any open `ReceiveStream`, write the data to it.
        """
        stream_id = event.stream_id

        log.debug("data received on stream %s: %s...", stream_id, event.data[:100])
        receive_stream = self.receive_streams.get(stream_id)
        if receive_stream is None:
            try:
                self.conn.reset_stream(stream_id, error_code=ErrorCodes.PROTOCOL_ERROR)
            except StreamClosedError:
                pass
            return

        receive_stream.write(event.data)
        self.conn.acknowledge_received_data(event.flow_controlled_length, stream_id)

    def window_updated(self, event):
        """ Called when the flow control window for a stream is changed.

        Any data waiting to be sent on the stream may fit in the window now.
        """
        log.debug("window updated, stream %s", event.stream_id)
        self.send_data(event.stream_id)

    def stream_ended(self, event):
        """ Called when an incoming stream ends.

        Close any `ReceiveStream` that was opened for this stream.
        """
        log.debug("stream ended, stream %s", event.stream_id)
        receive_stream = self.receive_streams.pop(event.stream_id, None)
        if receive_stream:
            receive_stream.close()

    def stream_reset(self, event):
        """ Called when an incoming stream is reset.

        Close any `ReceiveStream` that was opened for this stream.
        """
        log.debug("stream reset, stream %s", event.stream_id)
        receive_stream = self.receive_streams.pop(event.stream_id, None)
        if receive_stream:
            receive_stream.close()

    def settings_changed(self, event):
        log.debug("settings changed")

    def settings_acknowledged(self, event):
        log.debug("settings acknowledged")

    def trailers_received(self, event):
        log.debug("trailers received, stream %s", event.stream_id)
        receive_stream = self.receive_streams.get(event.stream_id)
        if not receive_stream:
            return

        receive_stream.trailers.set(*event.headers, from_wire=True)

    def connection_terminated(self, event):
        log.debug("connection terminated")
        self.run = False

    def send_headers(self, stream_id, immediate=False):
        """ Attempt to send any headers on a stream.

        Streams determine when headers should be sent. By default headers are not
        returned until there is also data ready to be sent. This can be overriddden
        with the `immediate` argument.
        """
        send_stream = self.send_streams.get(stream_id)

        if not send_stream:
            return

        headers = send_stream.headers_to_send(not immediate)
        if headers:
            self.conn.send_headers(stream_id, headers, end_stream=False)

    def send_data(self, stream_id):
        """ Attempt to send any pending data on a stream.

        Up to the current flow-control window size bytes may be sent.

        If the `SendStream` is exhausted (no more data to send), the stream is closed.
        """
        send_stream = self.send_streams.get(stream_id)

        if not send_stream:
            # window updates trigger sending of data, but can happen after a stream
            # has been completely sent
            return

        if not send_stream.headers_sent:
            # don't attempt to send any data until the headers have been sent
            return

        try:
            window_size = self.conn.local_flow_control_window(stream_id=stream_id)
            max_frame_size = self.conn.max_outbound_frame_size

            for chunk in send_stream.read(window_size, max_frame_size):
                log.debug("sending data on stream %s: %s...", stream_id, chunk[:100])

                self.conn.send_data(stream_id=stream_id, data=chunk)

        except StreamClosedError:
            return

        if send_stream.exhausted:
            log.debug("closing exhausted stream, stream %s", stream_id)
            self.end_stream(stream_id)

    def end_stream(self, stream_id):
        """ Close an outbound stream, sending any trailers.
        """
        send_stream = self.send_streams.pop(stream_id)

        try:
            trailers = send_stream.trailers_to_send()
            if trailers:
                self.conn.send_headers(stream_id, trailers, end_stream=True)
            else:
                self.conn.end_stream(stream_id)
        except StreamClosedError:
            pass


class ClientConnectionManager(ConnectionManager):
    """
    An object that manages a single HTTP/2 connection on a GRPC client.

    Extends the base `ConnectionManager` to make outbound GRPC requests.
    """

    def __init__(self, sock):
        super().__init__(sock, client_side=True)

        self.pending_requests = deque()

        self.counter = itertools.count(start=1, step=2)

    def on_iteration(self):
        """ On each iteration of the event loop, also initiate any pending requests.
        """
        self.send_pending_requests()
        super().on_iteration()

    def send_request(self, request_headers):
        """ Called by the client to invoke a GRPC method.

        Establish a `SendStream` to send the request payload and `ReceiveStream`
        for receiving the eventual response. `SendStream` and `ReceiveStream` are
        returned to the client for providing the request payload and iterating
        over the response.

        Invocations are queued and sent on the next iteration of the event loop.
        """
        stream_id = next(self.counter)

        request_stream = SendStream(stream_id)
        response_stream = ReceiveStream(stream_id)
        self.receive_streams[stream_id] = response_stream
        self.send_streams[stream_id] = request_stream

        request_stream.headers.set(*request_headers)

        self.pending_requests.append(stream_id)

        return request_stream, response_stream

    def response_received(self, event):
        """ Called when a response is received on a stream.

        If the headers contain an error, we should raise it here.
        """
        super().response_received(event)

        stream_id = event.stream_id
        response_stream = self.receive_streams.get(stream_id)
        if response_stream is None:
            self.conn.reset_stream(stream_id, error_code=ErrorCodes.PROTOCOL_ERROR)
            return

        headers = response_stream.headers

        if int(headers.get("grpc-status", 0)) > 0:
            error = GrpcError.from_headers(headers)
            response_stream.close(error)
            del self.receive_streams[stream_id]

    def trailers_received(self, event):
        """ Called when trailers are received on a stream.

        If the trailers contain an error, we should raise it here.
        """
        super().trailers_received(event)

        stream_id = event.stream_id
        response_stream = self.receive_streams.get(stream_id)
        if response_stream is None:
            self.conn.reset_stream(stream_id, error_code=ErrorCodes.PROTOCOL_ERROR)
            return

        trailers = response_stream.trailers

        if int(trailers.get("grpc-status", 0)) > 0:
            error = GrpcError.from_headers(trailers)
            response_stream.close(error)
            del self.receive_streams[stream_id]

    def send_pending_requests(self):
        """ Initiate requests for any pending invocations.

        Sends initial headers and any request data that is ready to be sent.
        """
        while self.pending_requests:
            stream_id = self.pending_requests.popleft()

            log.debug("initiating request, new stream %s", stream_id)

            # send headers immediately rather than waiting for data. this ensures
            # streams are established with increasing stream ids regardless of when
            # the request data is available
            self.send_headers(stream_id, immediate=True)
            self.send_data(stream_id)

    def send_data(self, stream_id):
        try:
            super().send_data(stream_id)
        except UnsupportedEncoding:

            response_stream = self.receive_streams[stream_id]
            request_stream = self.send_streams[stream_id]

            error = GrpcError(
                code=StatusCode.UNIMPLEMENTED,
                message="Algorithm not supported: {}".format(request_stream.encoding),
            )
            response_stream.close(error)
            request_stream.close()

            del self.receive_streams[stream_id]
            del self.send_streams[stream_id]


class ServerConnectionManager(ConnectionManager):
    """
    An object that manages a single HTTP/2 connection on a GRPC server.

    Extends the base `ConnectionManager` to handle incoming GRPC requests.
    """

    def __init__(self, sock, handle_request):
        super().__init__(sock, client_side=False)
        self.handle_request = handle_request

    def request_received(self, event):
        """ Receive a GRPC request and pass it to the GrpcServer to fire any
        appropriate entrypoint.

        Establish a `ReceiveStream` to receive the request payload and `SendStream`
        for sending the eventual response.
        """
        super().request_received(event)

        stream_id = event.stream_id

        request_stream = ReceiveStream(stream_id)
        response_stream = SendStream(stream_id)
        self.receive_streams[stream_id] = request_stream
        self.send_streams[stream_id] = response_stream

        request_stream.headers.set(*event.headers, from_wire=True)

        compression = select_algorithm(
            request_stream.headers.get("grpc-accept-encoding"),
            request_stream.headers.get("grpc-encoding"),
        )

        try:
            response_stream.headers.set(
                (":status", "200"),
                ("content-type", "application/grpc+proto"),
                ("grpc-accept-encoding", ",".join(SUPPORTED_ENCODINGS)),
                # TODO support server changing compression later
                ("grpc-encoding", compression),
            )
            response_stream.trailers.set(("grpc-status", "0"))
            self.handle_request(request_stream, response_stream)

        except GrpcError as error:
            response_stream.trailers.set((":status", "200"), *error.as_headers())
            self.end_stream(stream_id)

    def send_data(self, stream_id):
        try:
            super().send_data(stream_id)
        except GrpcError as error:
            send_stream = self.send_streams.get(stream_id)
            send_stream.trailers.set(*error.as_headers())
            self.end_stream(stream_id)



# -*- coding: utf-8 -*-
from collections import OrderedDict


buckets = OrderedDict(
    [
        (0.000000001, "n"),
        (0.000001, "u"),
        (0.001, "m"),
        (1, "S"),
        (60, "M"),
        (3600, "H"),
    ]
)
inverse_buckets = {unit: value for value, unit in buckets.items()}


def unbucket_timeout(value):
    unit = value[-1]
    count = int(value[:-1])
    multiplier = inverse_buckets[unit]
    return count * multiplier


def bucket_timeout(value):

    periods = list(buckets)
    bucket_period = periods.pop(0)

    for period in periods:
        if value // period > 1:
            bucket_period = period
        else:
            break
    return "{}{}".format(round(value / bucket_period), buckets[bucket_period])



# -*- coding: utf-8 -*-
import base64


def is_grpc_header(name):
    return name.startswith("grpc-")


def is_pseudo_header(name):
    return name.startswith(":")


def is_http2_header(name):
    return name in ("te", "content-type", "user-agent", "accept-encoding")


def filter_headers_for_application(headers):
    def include(header):
        name, _ = header
        if is_pseudo_header(name) or is_grpc_header(name) or is_http2_header(name):
            return False
        return True

    return list(filter(include, headers))


def sort_headers_for_wire(headers):
    def weight(header):
        name, _ = header
        if is_pseudo_header(name):
            return 0
        if is_http2_header(name):
            return 1
        if is_grpc_header(name):
            return 2
        return 3

    return sorted(headers, key=weight)


def check_encoded(headers):
    for name, value in headers:
        assert isinstance(name, bytes)
        assert isinstance(value, bytes)


def check_decoded(headers):
    for name, value in headers:
        assert not isinstance(name, bytes)
        if name.endswith("-bin"):
            assert isinstance(value, bytes)
        else:
            assert not isinstance(value, bytes)


def decode_header(header):
    name, value = header
    name = name.decode("utf-8")
    if name.endswith("-bin"):
        value = base64.b64decode(
            value + b"=="
        )  # add padding, per https://stackoverflow.com/a/49459036/128749
    else:
        value = value.decode("utf-8")
    return name, value


def encode_header(header):
    name, value = header
    name = name.encode("utf-8")
    if name.endswith(b"-bin"):
        value = base64.b64encode(value)
    else:
        value = value.encode("utf-8")
    return name, value


def comma_join(values):
    if all(map(lambda item: isinstance(item, bytes), values)):
        separator = b","
    else:
        separator = ","
    return separator.join(values)


class HeaderManager:
    def __init__(self):
        """
        Object for managing headers (or trailers). Headers are encoded before
        transmitting over the wire and stored in their non-encoded form.
        """
        self.data = []

    @staticmethod
    def decode(headers):
        return list(map(decode_header, headers))

    @staticmethod
    def encode(headers):
        return list(map(encode_header, headers))

    def get(self, name, default=None):
        """ Get a header by `name`.

        Joins duplicate headers into a comma-separated string of values.
        """
        matches = []
        for key, value in self.data:
            if key == name:
                matches.append(value)
        if not matches:
            return default
        return comma_join(matches)

    def set(self, *headers, from_wire=False):
        """ Set headers.

        Overwrites any existing header with the same name. Optionally decodes
        the headers first.
        """
        # XXX why isn't this clobbered?
        if from_wire:
            headers = self.decode(headers)
        check_decoded(headers)

        # clear existing headers with these names
        to_clear = dict(headers).keys()
        self.data = [(key, value) for (key, value) in self.data if key not in to_clear]

        self.data.extend(headers)

    def append(self, *headers, from_wire=False):
        """ Add new headers.

        Preserves and appends to any existing header with the same name. Optionally
        decodes the headers first.
        """
        if from_wire:
            headers = self.decode(headers)
        check_decoded(headers)
        self.data.extend(headers)

    def __len__(self):
        return len(self.data)

    @property
    def for_wire(self):
        """ A sorted list of encoded headers for transmitting over the wire.
        """
        return self.encode(sort_headers_for_wire(self.data))

    @property
    def for_application(self):
        """ A filtered list of headers for use by the application.
        """
        return filter_headers_for_application(self.data)



# -*- coding: utf-8 -*-
import struct
from queue import Empty, Queue

from nameko_grpc.compression import compress, decompress
from nameko_grpc.errors import GrpcError
from nameko_grpc.headers import HeaderManager


HEADER_LENGTH = 5

STREAM_END = object()


class ByteBuffer:
    def __init__(self):
        self.bytes = bytearray()

    def peek(self, view=None):
        if view is None:
            view = slice(0, len(self.bytes))
        return self.bytes[view]

    def discard(self, max_bytes=None):
        self.read(max_bytes)

    def read(self, max_bytes=None):
        if max_bytes is None:
            max_bytes = len(self.bytes)
        length = min(max_bytes, len(self.bytes))
        data = self.bytes[:length]
        self.bytes = self.bytes[length:]
        return data

    def write(self, data):
        self.bytes.extend(data)

    def empty(self):
        return len(self.bytes) == 0

    def __len__(self):
        return len(self.bytes)


class StreamBase:
    def __init__(self, stream_id):
        self.stream_id = stream_id

        self.headers = HeaderManager()
        self.trailers = HeaderManager()

        self.queue = Queue()
        self.buffer = ByteBuffer()
        self.closed = False

    @property
    def exhausted(self):
        """ A stream is exhausted if it is closed and there are no more messages to be
        consumed or bytes to be read.
        """
        return self.closed and self.queue.empty() and self.buffer.empty()

    def close(self, error=None):
        """ Close this stream, preventing further messages or data to be added.

        If closed with an error, the error will be raised when reading
        or consuming from this stream.

        This method may be called after a stream is already closed. This happens
        in race conditions between timeout threads, connection teardown, and the
        natural termination of streams.

        SendStreams have an additional race condition beteen the end of the iterator
        and the StreamEnded event received from the remote side.

        An error is only raised if the first invocation happened due to an error.
        """
        if self.closed:
            return

        if error:
            assert isinstance(error, GrpcError)

        self.closed = True
        self.queue.put(error or STREAM_END)


class ReceiveStream(StreamBase):
    """ An HTTP2 stream that receives data as bytes to be iterated over as GRPC
    messages.
    """

    def write(self, data):
        """ Write data to this stream, separating it into message-sized chunks.
        """
        if self.closed:
            return

        self.buffer.write(data)
        while True:

            if len(self.buffer) < HEADER_LENGTH:
                break

            compressed_flag = struct.unpack("?", self.buffer.peek(slice(0, 1)))[0]
            message_length = struct.unpack(">I", self.buffer.peek(slice(1, 5)))[0]

            if len(self.buffer) < HEADER_LENGTH + message_length:
                break

            self.buffer.discard(HEADER_LENGTH)
            message_data = bytes(self.buffer.read(message_length))
            self.queue.put((compressed_flag, message_data))

    def consume(self, message_type):
        """ Consume the data in this stream by yielding `message_type` messages,
        or raising if the stream was closed with an error.
        """
        while True:
            item = self.queue.get()
            if isinstance(item, GrpcError):
                raise item
            elif item is STREAM_END:
                break

            compressed, message_data = item
            if compressed:
                message_data = decompress(message_data)

            message = message_type()
            message.ParseFromString(message_data)

            yield message


class SendStream(StreamBase):
    """ An HTTP2 stream that receives data as GRPC messages to be read as chunks of
    bytes.
    """

    def __init__(self, *args, **kwargs):
        self.headers_sent = False
        super().__init__(*args, **kwargs)

    @property
    def encoding(self):
        return self.headers.get("grpc-encoding")

    def populate(self, iterable):
        """ Populate this stream with an iterable of messages.
        """
        for item in iterable:
            if self.closed:
                return
            self.queue.put(item)
        self.close()

    def headers_to_send(self, defer_until_data=True):
        """ Return any headers to be sent with this stream.

        Headers may only be transmitted before any data is sent.
        This state is maintained by only returning headers from this method once.

        When `defer_until_data` is true, no headers are returned until there is at least
        one message ready to be sent. This allows for header values to be changed until
        the last possible moment (enabling the server to change encoding, for example).
        """
        if self.headers_sent or len(self.headers) == 0:
            return False

        if defer_until_data and self.queue.empty():
            return False

        self.headers_sent = True
        return self.headers.for_wire

    def trailers_to_send(self):
        """ Return any trailers to be sent after this stream.
        """
        if len(self.trailers) == 0:
            return False

        return self.trailers.for_wire

    def flush_queue_to_buffer(self):
        """ Write the bytes from any messages in the queue to the buffer.
        """
        while True:
            try:
                message = self.queue.get_nowait()
            except Empty:
                break

            # any error should be raised immediately
            if isinstance(message, GrpcError):
                raise message

            # add the bytes from the message to the buffer
            if message and message != STREAM_END:
                body = message.SerializeToString()
                compressed, body = compress(body, self.encoding)

                data = (
                    struct.pack("?", compressed) + struct.pack(">I", len(body)) + body
                )
                self.buffer.write(data)

    def read(self, max_bytes, chunk_size):
        """ Read up to `max_bytes` from the stream, yielding up to `chunk_size`
        bytes at a time.
        """
        sent = 0

        while sent < max_bytes:

            # ensure any messages enqueued during reading are flushed
            self.flush_queue_to_buffer()

            # chunk data out of buffer
            max_read = min(chunk_size, max_bytes - sent)
            chunk = self.buffer.read(max_read)
            if not chunk:
                break  # buffer is empty
            sent += len(chunk)
            yield chunk



# -*- coding: utf-8 -*-
import json

import pytest
from nameko.rpc import rpc
from nameko.standalone.rpc import ServiceRpcProxy

from nameko_grpc.dependency_provider import GrpcProxy


@pytest.fixture(autouse=True)
def stubs(load_stubs):
    return load_stubs("advanced")


@pytest.fixture(autouse=True)
def protobufs(load_protobufs):
    return load_protobufs("advanced")


class TestContextData:
    @pytest.fixture(params=["server=nameko"])
    def server_type(self, request):
        # only nameko server supports the features in this test
        return request.param[7:]

    @pytest.fixture
    def grpc_server(self, start_server):
        start_server("advanced")

    @pytest.fixture
    def grpc_client(self, start_client, grpc_server):
        return start_client("advanced")

    @pytest.fixture
    def amqprpc_service(
        self, container_factory, rabbit_config, grpc_server, protobufs, stubs, grpc_port
    ):
        class AmqpRpcService:
            name = "amqp"

            advanced_grpc = GrpcProxy(
                "//127.0.0.1:{}".format(grpc_port), stubs.advancedStub
            )

            @rpc
            def proxy(self):
                response = self.advanced_grpc.unary_unary(
                    protobufs.SimpleRequest(value="A"),
                    metadata=[("x", "X"), ("x-bin", b"\x0a\x0b\x0a\x0b\x0a\x0b")],
                )
                return response.message

        container = container_factory(AmqpRpcService, rabbit_config)
        container.start()

    @pytest.mark.parametrize(
        "metadata,expected_key,expected_value",
        [
            # simple
            ([("a", "A")], "a", "A"),
            # duplicate names
            ([("a", "A1"), ("a", "A2")], "a", ["A1", "A2"]),
            # # binary
            ([("a-bin", b"\x0a\x0b\x0a\x0b\x0a\x0b")], "a-bin", "CgsKCwoL"),
            # # duplicate binary
            (
                [
                    ("a-bin", b"\x01\x01\x01\x01\x01\x01"),
                    ("a-bin", b"\x02\x02\x02\x02\x02\x02"),
                ],
                "a-bin",
                ["AQEBAQEB", "AgICAgIC"],
            ),
        ],
    )
    def test_request_metadata_available_as_context_data(
        self, grpc_client, protobufs, metadata, expected_key, expected_value
    ):
        response = grpc_client.unary_unary(
            protobufs.SimpleRequest(value="A"), metadata=metadata
        )
        assert json.loads(response.message)[expected_key] == expected_value

    @pytest.mark.usefixtures("predictable_call_ids")
    def test_dependency_provider_includes_context_data_in_grpc_request_metadata(
        self, amqprpc_service, rabbit_config
    ):
        with ServiceRpcProxy("amqp", context_data={"a": "A"}) as proxy_rpc:

            context_data = json.loads(proxy_rpc.proxy())

            # added by RPC client, propagated by GRPC
            assert context_data["a"] == "A"

            # added by GRPC client as metadata
            assert context_data["x"] == "X"
            assert context_data["x-bin"] == "CgsKCwoL"  # base64-encoded for response

            # NOTE adding binary objects to context_data will cause problems for
            # extensions that can't serialize them (e.g. AMQP entrypoints in their
            # default configuration); this is a general problem when mixing extensions.
            # opportunity to improve this when nameko grows its own serialization layer

            # call id stack is propagated
            assert context_data["call_id_stack"] == [
                "standalone_rpc_client.0.0",
                "amqp.proxy.1",
                "advanced.unary_unary.2",
            ]



# -*- coding: utf-8 -*-
import gzip
import zlib
from collections import OrderedDict


ENCODERS = OrderedDict(
    [
        ("deflate", {"compress": zlib.compress, "decompress": zlib.decompress}),
        ("gzip", {"compress": gzip.compress, "decompress": gzip.decompress}),
    ]
)
SUPPORTED_ENCODINGS = tuple(ENCODERS.keys()) + ("identity",)


class UnsupportedEncoding(Exception):
    pass


def decompress(data):
    for algorithm in ENCODERS.values():
        try:
            return algorithm["decompress"](data)
        except Exception:
            pass
    raise UnsupportedEncoding(
        "Could not decompress data. Supported algorithms: {})".format(
            ", ".join(SUPPORTED_ENCODINGS)
        )
    )


def compress(data, encoding):  # level?
    if encoding == "identity":
        return False, data
    if encoding in ENCODERS:
        # TODO: heuristic to decide whether we _should_ compress or not
        return True, ENCODERS[encoding]["compress"](data)
    raise UnsupportedEncoding(encoding)


def select_algorithm(acceptable_encodings, preferred_encoding):
    if preferred_encoding and preferred_encoding in SUPPORTED_ENCODINGS:
        return preferred_encoding
    for encoding in SUPPORTED_ENCODINGS:
        if encoding in acceptable_encodings:
            return encoding
    raise UnsupportedEncoding(
        "No supported algorithm acceptable: {} (supported: {})".format(
            acceptable_encodings, ", ".join(SUPPORTED_ENCODINGS)
        )
    )



# -*- coding: utf-8 -*-
import json
import random
import string

import pytest

from nameko_grpc.constants import Cardinality


@pytest.mark.equivalence
class TestConcurrency:
    def test_unary_unary(self, client, protobufs, instrumented, client_type):

        stash_metadata = json.dumps([instrumented.path, Cardinality.UNARY_UNARY.value])

        futures = []
        for letter in string.ascii_uppercase:
            futures.append(
                client.unary_unary.future(
                    protobufs.ExampleRequest(value=letter),
                    metadata=[("stash", stash_metadata)],
                )
            )

        for index, future in enumerate(futures):
            assert future.result().message == string.ascii_uppercase[index]

        # verify messages from concurrent requests are interleaved
        # there is a 1/26! chance of concurrent requests being handled in order
        captured_requests = list(instrumented.requests())
        assert len(captured_requests) == 26
        assert [req.value for req in captured_requests] != string.ascii_uppercase

    def test_unary_stream(self, client, protobufs, instrumented, client_type):

        stash_metadata = json.dumps([instrumented.path, Cardinality.UNARY_STREAM.value])

        futures = []
        for letter in string.ascii_uppercase:
            futures.append(
                client.unary_stream.future(
                    protobufs.ExampleRequest(value=letter, response_count=2),
                    metadata=[("stash", stash_metadata)],
                )
            )

        for index, future in enumerate(futures):
            result = list(future.result())
            responses = [(response.message, response.seqno) for response in result]
            assert responses == [
                (string.ascii_uppercase[index], 1),
                (string.ascii_uppercase[index], 2),
            ]

        # verify messages from concurrent requests are interleaved
        # there is a 1/26! chance of concurrent requests being handled in order
        captured_requests = list(instrumented.requests())
        assert len(captured_requests) == 26
        assert [req.value for req in captured_requests] != string.ascii_uppercase

    def test_stream_unary(
        self, client, protobufs, instrumented, client_type, server_type
    ):

        stash_metadata = json.dumps([instrumented.path, Cardinality.STREAM_UNARY.value])

        def generate_requests(values):
            for value in values:
                yield protobufs.ExampleRequest(value=value)

        futures = []
        for index in range(26):
            if index % 2 == 0:
                values = string.ascii_uppercase
            else:
                values = string.ascii_lowercase
            futures.append(
                client.stream_unary.future(
                    generate_requests(values), metadata=[("stash", stash_metadata)]
                )
            )

        for index, future in enumerate(futures):
            if index % 2 == 0:
                assert future.result().message == ",".join(string.ascii_uppercase)
            else:
                assert future.result().message == ",".join(string.ascii_lowercase)

        # verify messages from concurrent requests are interleaved
        # there is a 1/626! chance of concurrent requests being handled in order,
        # just check the first 26.
        captured_requests = list(instrumented.requests())
        assert len(captured_requests) == 26 * 26
        assert [req.value for req in captured_requests[:26]] != string.ascii_uppercase

    def test_stream_stream(self, client, protobufs, instrumented, client_type):

        stash_metadata = json.dumps(
            [instrumented.path, Cardinality.STREAM_STREAM.value]
        )

        def generate_requests(values):
            for value in values:
                yield protobufs.ExampleRequest(value=value)

        futures = []
        for index in range(26):
            if index % 2 == 0:
                values = string.ascii_uppercase
            else:
                values = string.ascii_lowercase
            futures.append(
                client.stream_stream.future(
                    generate_requests(values), metadata=[("stash", stash_metadata)]
                )
            )

        for index, future in enumerate(futures):
            result = list(future.result())
            responses = [(response.seqno, response.message) for response in result]
            if index % 2 == 0:
                expected = list(enumerate(string.ascii_uppercase, 1))
            else:
                expected = list(enumerate(string.ascii_lowercase, 1))
            assert responses == expected

        # verify messages from concurrent requests are interleaved
        # there is a 1/626! chance of concurrent requests being handled in order,
        # just check the first 26.
        captured_requests = list(instrumented.requests())
        assert len(captured_requests) == 26 * 26
        assert [req.value for req in captured_requests[:26]] != string.ascii_uppercase


@pytest.mark.equivalence
class TestMultipleClients:
    def test_unary_unary(self, start_client, server, protobufs):

        futures = []
        number_of_clients = 5

        for index in range(number_of_clients):
            client = start_client("example")
            response_future = client.unary_unary.future(
                protobufs.ExampleRequest(value=string.ascii_uppercase[index])
            )
            futures.append(response_future)

        for index, future in enumerate(futures):
            response = future.result()
            assert response.message == string.ascii_uppercase[index]

    def test_unary_stream(self, start_client, server, protobufs):

        futures = []
        number_of_clients = 5

        for index in range(number_of_clients):
            client = start_client("example")
            responses_future = client.unary_stream.future(
                protobufs.ExampleRequest(
                    value=string.ascii_uppercase[index], response_count=2
                )
            )
            futures.append(responses_future)

        for index, future in enumerate(futures):
            responses = future.result()
            assert [(response.message, response.seqno) for response in responses] == [
                (string.ascii_uppercase[index], 1),
                (string.ascii_uppercase[index], 2),
            ]

    def test_stream_unary(self, start_client, server, protobufs):

        number_of_clients = 5

        def shuffled(string):
            chars = list(string)
            random.shuffle(chars)
            return chars

        streams = [shuffled(string.ascii_uppercase) for _ in range(number_of_clients)]

        def generate_requests(values):
            for value in values:
                yield protobufs.ExampleRequest(value=value)

        futures = []

        for index in range(number_of_clients):
            client = start_client("example")
            response_future = client.stream_unary.future(
                generate_requests(streams[index])
            )
            futures.append(response_future)

        for index, future in enumerate(futures):
            response = future.result()
            assert response.message == ",".join(streams[index])

    def test_stream_stream(self, start_client, server, protobufs):

        number_of_clients = 5

        def shuffled(string):
            chars = list(string)
            random.shuffle(chars)
            return chars

        streams = [shuffled(string.ascii_uppercase) for _ in range(number_of_clients)]

        def generate_requests(values):
            for value in values:
                yield protobufs.ExampleRequest(value=value)

        futures = []

        for index in range(number_of_clients):
            client = start_client("example")
            responses_future = client.stream_stream.future(
                generate_requests(streams[index])
            )
            futures.append(responses_future)

        for index, future in enumerate(futures):
            responses = future.result()

            received = [(response.seqno, response.message) for response in responses]
            assert received == list(enumerate(streams[index], 1))



# -*- coding: utf-8 -*-
import json

from nameko_grpc.errors import STATUS_CODE_ENUM_TO_INT_MAP


METADATA_PREFIX = "x-nameko-"


def encode_value(value):
    # TODO use nameko.serialization
    return json.dumps(value)


def decode_value(value):
    # TODO use nameko.serialization
    return json.loads(value)


def metadata_from_context_data(data):
    """ Utility function transforming a `data` dictionary into metadata suitable
    for a GRPC stream.
    """
    metadata = []
    for key, value in data.items():
        name = "{}{}".format(METADATA_PREFIX, key)
        value = encode_value(value)
        metadata.append((name, value))
    return metadata


def context_data_from_metadata(metadata):
    """ Utility function transforming `metadata` into a context data dictionary.

    Metadata may have been encoded at the client by `metadata_from_context_data`, or
    it may be "normal" GRPC metadata. In this case, duplicate values are allowed;
    they become a list in the context data.
    """
    data = {}

    for name, value in metadata:
        if name.startswith(METADATA_PREFIX):
            _, key = name.split(METADATA_PREFIX, 1)
            data[key] = decode_value(value)
        else:
            if name in data:
                try:
                    data[name].append(value)
                except AttributeError:
                    data[name] = [data[name], value]
            else:
                data[name] = value

    return data


class GrpcContext:
    """ Context object passed to GRPC methods.

    Gives access to request metadata and allows response metadata to be set.
    """

    def __init__(self, request_stream, response_stream):
        self.request_stream = request_stream
        self.response_stream = response_stream

    def set_code(self, code):
        self.response_stream.trailers.set(
            ("grpc-status", str(STATUS_CODE_ENUM_TO_INT_MAP[code]))
        )

    def set_message(self, message):
        self.response_stream.trailers.set(("grpc-message", message))

    def invocation_metadata(self):
        return self.request_stream.headers.for_application

    def send_initial_metadata(self, metadata):
        self.response_stream.headers.set(*metadata)

    def set_trailing_metadata(self, metadata):
        self.response_stream.trailers.set(*metadata)



# -*- coding: utf-8 -*-
import base64
import json

from nameko.extensions import DependencyProvider

from nameko_grpc.entrypoint import Grpc

import advanced_pb2_grpc
from advanced_pb2 import SimpleReply


grpc = Grpc.implementing(advanced_pb2_grpc.advancedStub)


class BytesEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, bytes):
            return base64.b64encode(obj).decode("utf-8")
        return json.JSONEncoder.default(self, obj)


class ContextData(DependencyProvider):
    def get_dependency(self, worker_ctx):
        return worker_ctx.context_data


class advanced:
    name = "advanced"

    context_data = ContextData()

    @grpc
    def unary_unary(self, request, context):
        message = json.dumps(self.context_data, cls=BytesEncoder)
        return SimpleReply(message=message)



# -*- coding: utf-8 -*-
import pytest


@pytest.mark.equivalence
class TestStandard:
    def test_unary_unary(self, client, protobufs):
        response = client.unary_unary(protobufs.ExampleRequest(value="A"))
        assert response.message == "A"

    def test_unary_stream(self, client, protobufs):
        responses = client.unary_stream(
            protobufs.ExampleRequest(value="A", response_count=2)
        )
        assert [(response.message, response.seqno) for response in responses] == [
            ("A", 1),
            ("A", 2),
        ]

    def test_stream_unary(self, client, protobufs):
        def generate_requests():
            for value in ["A", "B"]:
                yield protobufs.ExampleRequest(value=value)

        response = client.stream_unary(generate_requests())
        assert response.message == "A,B"

    def test_stream_stream(self, client, protobufs):
        def generate_requests():
            for value in ["A", "B"]:
                yield protobufs.ExampleRequest(value=value)

        responses = client.stream_stream(generate_requests())
        assert [(response.message, response.seqno) for response in responses] == [
            ("A", 1),
            ("B", 2),
        ]


@pytest.mark.equivalence
class TestLarge:
    def test_large_unary_request(self, client, protobufs):
        response = client.unary_unary(
            protobufs.ExampleRequest(value="A", blob="X" * 20000)
        )
        assert response.message == "A"

    def test_large_unary_response(self, client, protobufs):
        multiplier = 20000
        response = client.unary_unary(
            protobufs.ExampleRequest(value="A", multiplier=multiplier)
        )
        assert response.message == "A" * multiplier

    def test_large_streaming_request(self, client, protobufs):
        def generate_requests():
            for value in ["A", "B"]:
                yield protobufs.ExampleRequest(value=value, blob="X" * 20000)

        response = client.stream_unary(generate_requests())
        assert response.message == "A,B"

    def test_large_streaming_response(self, client, protobufs):

        multiplier = 20000
        responses = client.unary_stream(
            protobufs.ExampleRequest(value="A", multiplier=multiplier, response_count=2)
        )
        assert [(response.message, response.seqno) for response in responses] == [
            ("A" * multiplier, 1),
            ("A" * multiplier, 2),
        ]


@pytest.mark.equivalence
class TestFuture:
    def test_unary_unary(self, client, protobufs):
        response_future = client.unary_unary.future(protobufs.ExampleRequest(value="A"))
        response = response_future.result()
        assert response.message == "A"

    def test_unary_stream(self, client, protobufs):
        responses_future = client.unary_stream.future(
            protobufs.ExampleRequest(value="A", response_count=2)
        )
        responses = responses_future.result()
        assert [(response.message, response.seqno) for response in responses] == [
            ("A", 1),
            ("A", 2),
        ]

    def test_stream_unary(self, client, protobufs):
        def generate_requests():
            for value in ["A", "B"]:
                yield protobufs.ExampleRequest(value=value)

        response_future = client.stream_unary.future(generate_requests())
        response = response_future.result()
        assert response.message == "A,B"

    def test_stream_stream(self, client, protobufs):
        def generate_requests():
            for value in ["A", "B"]:
                yield protobufs.ExampleRequest(value=value)

        responses_future = client.stream_stream.future(generate_requests())
        responses = responses_future.result()
        assert [(response.message, response.seqno) for response in responses] == [
            ("A", 1),
            ("B", 2),
        ]



# -*- coding: utf-8 -*-
import sys
import threading
from importlib import import_module

import grpc
import zmq
from grpc._cython.cygrpc import CompressionAlgorithm, CompressionLevel

from nameko_grpc.constants import Cardinality
from nameko_grpc.errors import GrpcError

from helpers import Command, RemoteClientTransport, status_from_metadata


def execute(command, stub):
    method = getattr(stub, command.method_name)

    request = command.get_request()

    compression = command.kwargs.pop("compression", None)
    if compression:
        command.kwargs["metadata"] = list(command.kwargs.get("metadata", [])) + [
            ("grpc-internal-encoding-request", compression)
        ]

    response_metadata = {}

    try:
        if command.cardinality in (Cardinality.STREAM_UNARY, Cardinality.UNARY_UNARY):
            response_future = method.future(request, **command.kwargs)
            response_metadata["code"] = response_future.code()
            response_metadata["details"] = response_future.details()
            response_metadata["initial_metadata"] = list(
                map(tuple, response_future.initial_metadata())
            )
            response_metadata["trailing_metadata"] = list(
                map(tuple, response_future.trailing_metadata())
            )
            response = response_future.result()
        else:
            # .future() interface for RPCs with STREAM responses not supported
            response = method(request, **command.kwargs)
    except grpc.RpcError as exc:
        state = exc._state
        response_metadata["code"] = state.code
        response_metadata["details"] = state.details

        response = GrpcError(
            state.code, state.details, status_from_metadata(state.trailing_metadata)
        )

    command.send_response(response)
    command.send_metadata(response_metadata)


if __name__ == "__main__":

    port = sys.argv[1]
    secure = sys.argv[2]

    proto_name = sys.argv[3]
    service_name = sys.argv[4]

    compression_algorithm = sys.argv[5]
    compression_level = sys.argv[6]

    grpc_module = import_module("{}_pb2_grpc".format(proto_name))
    stub_cls = getattr(grpc_module, "{}Stub".format(service_name))

    zmq_port = sys.argv[7]

    channel_options = [
        (
            "grpc.default_compression_algorithm",
            getattr(CompressionAlgorithm, compression_algorithm),
        ),
        (
            "grpc.default_compression_level",
            getattr(CompressionLevel, compression_level),
        ),
    ]

    if secure == "secure":
        with open("test/certs/server.crt", "rb") as f:
            creds = grpc.ssl_channel_credentials(f.read())
        channel = grpc.secure_channel(
            "localhost:{}".format(port), creds, options=channel_options
        )
    else:
        channel = grpc.insecure_channel(
            "127.0.0.1:{}".format(port), options=channel_options
        )
    stub = stub_cls(channel)

    transport = RemoteClientTransport.connect(
        zmq.Context(), zmq.REP, "tcp://127.0.0.1:{}".format(zmq_port)
    )

    for command in Command.retrieve_commands(transport):
        threading.Thread(target=execute, args=(command, stub)).start()



# -*- coding: utf-8 -*-
from grpc_status import rpc_status

from nameko_grpc.errors import StatusCode, make_status

from helpers import extract_metadata, instrumented, maybe_echo_metadata, maybe_sleep

import example_pb2_grpc
from example_pb2 import ExampleReply


class Error(Exception):
    pass


class example(example_pb2_grpc.exampleServicer):
    @instrumented
    def unary_unary(self, request, context):
        metadata = extract_metadata(context)
        maybe_echo_metadata(context)
        maybe_sleep(request)
        message = request.value * (request.multiplier or 1)
        return ExampleReply(message=message, metadata=metadata)

    @instrumented
    def unary_stream(self, request, context):
        metadata = extract_metadata(context)
        maybe_echo_metadata(context)
        message = request.value * (request.multiplier or 1)
        for i in range(request.response_count):
            maybe_sleep(request)
            yield ExampleReply(message=message, seqno=i + 1, metadata=metadata)

    @instrumented
    def stream_unary(self, request, context):
        metadata = extract_metadata(context)
        maybe_echo_metadata(context)
        messages = []
        for index, req in enumerate(request):
            maybe_sleep(req)
            message = req.value * (req.multiplier or 1)
            messages.append(message)

        return ExampleReply(message=",".join(messages), metadata=metadata)

    @instrumented
    def stream_stream(self, request, context):
        metadata = extract_metadata(context)
        maybe_echo_metadata(context)
        for index, req in enumerate(request):
            maybe_sleep(req)
            message = req.value * (req.multiplier or 1)
            yield ExampleReply(message=message, seqno=index + 1, metadata=metadata)

    @instrumented
    def unary_error(self, request, context):
        maybe_echo_metadata(context)
        maybe_sleep(request)
        raise Error("boom")

    @instrumented
    def unary_error_via_context(self, request, context):
        # using rich status to test compatibility with
        # https://grpc.github.io/grpc/python/grpc_status.html
        status = make_status(code=StatusCode.UNAUTHENTICATED, message="Not allowed!",)
        context.abort_with_status(rpc_status.to_status(status))

    @instrumented
    def stream_error(self, request, context):
        metadata = extract_metadata(context)
        maybe_echo_metadata(context)
        message = request.value * (request.multiplier or 1)
        for i in range(request.response_count):
            maybe_sleep(request)
            # raise on the last message
            if i == request.response_count - 1:
                raise Error("boom")
            yield ExampleReply(message=message, seqno=i + 1, metadata=metadata)

    @instrumented
    def stream_error_via_context(self, request, context):
        message = request.value * (request.multiplier or 1)
        for i in range(request.response_count):
            maybe_sleep(request)
            # raise on the last message
            if i == request.response_count - 1:
                # using rich status to test compatibility with
                # https://grpc.github.io/grpc/python/grpc_status.html
                status = make_status(
                    code=StatusCode.RESOURCE_EXHAUSTED, message="Out of tokens!"
                )
                context.abort_with_status(rpc_status.to_status(status))

            yield ExampleReply(message=message, seqno=i + 1)



# -*- coding: utf-8 -*-
import json
import re
import string
import time

import pytest
from grpc import StatusCode
from nameko import config

from nameko_grpc.constants import Cardinality
from nameko_grpc.errors import (
    STATUS_CODE_ENUM_TO_INT_MAP,
    GrpcError,
    register_exception_handler,
)

from google.protobuf.any_pb2 import Any
from google.rpc.error_details_pb2 import DebugInfo
from google.rpc.status_pb2 import Status


@pytest.mark.equivalence
class TestMethodNotFound:
    @pytest.fixture(autouse=True)
    def unregister_grpc_method(self, stubs):
        with open(stubs.__file__) as fh:
            original_service = fh.read()

        pattern = re.compile(r"'not_found': grpc.\w+\(.*?\),", re.DOTALL)
        modified_service = re.sub(pattern, "", original_service)

        with open(stubs.__file__, "w") as fh:
            fh.write(modified_service)

        yield

        with open(stubs.__file__, "w") as fh:
            fh.write(original_service)

    def test_method_not_found(self, client, protobufs):
        with pytest.raises(GrpcError) as error:
            client.not_found(protobufs.ExampleRequest(value="hello"))
        assert error.value.code == StatusCode.UNIMPLEMENTED
        assert error.value.message == "Method not found!"


@pytest.mark.equivalence
class TestDeadlineExceededAtClient:
    @pytest.fixture
    def protobufs(self, compile_proto, spec_dir):
        protobufs, _ = compile_proto("example")
        return protobufs

    def test_timeout_before_any_result(self, client, protobufs):
        with pytest.raises(GrpcError) as error:
            client.unary_unary(
                protobufs.ExampleRequest(value="A", delay=1000),
                timeout=0.05,  # XXX fails when too fast; need protection
            )
        assert error.value.code == StatusCode.DEADLINE_EXCEEDED
        assert error.value.message == "Deadline Exceeded"

    def test_timeout_while_streaming_request(self, client, protobufs):
        def generate_requests(values):
            for value in values:
                time.sleep(0.01)
                yield protobufs.ExampleRequest(value=value)

        with pytest.raises(GrpcError) as error:
            client.stream_unary(generate_requests(string.ascii_uppercase), timeout=0.05)
        assert error.value.code == StatusCode.DEADLINE_EXCEEDED
        assert error.value.message == "Deadline Exceeded"

    def test_timeout_while_streaming_result(self, client, protobufs):

        res = client.unary_stream(
            protobufs.ExampleRequest(value="A", delay=10, response_count=10),
            timeout=0.05,
        )
        with pytest.raises(GrpcError) as error:
            list(res)

        assert error.value.code == StatusCode.DEADLINE_EXCEEDED
        assert error.value.message == "Deadline Exceeded"


@pytest.mark.equivalence
class TestDeadlineExceededAtServer:
    @pytest.fixture
    def protobufs(self, compile_proto, spec_dir):
        protobufs, _ = compile_proto("example")
        return protobufs

    def test_timeout_while_streaming_request(self, client, protobufs, instrumented):

        stash_metadata = json.dumps([instrumented.path, Cardinality.STREAM_UNARY.value])

        def generate_requests(values):
            for value in values:
                time.sleep(0.01)
                yield protobufs.ExampleRequest(value=value)

        with pytest.raises(GrpcError) as error:
            client.stream_unary(
                generate_requests(string.ascii_uppercase),
                timeout=0.05,
                metadata=[("stash", stash_metadata)],
            )
        assert error.value.code == StatusCode.DEADLINE_EXCEEDED
        assert error.value.message == "Deadline Exceeded"

        # server should not have recieved all the requests
        captured_requests = list(instrumented.requests())
        assert len(captured_requests) < len(string.ascii_uppercase)

    def test_timeout_while_streaming_response(self, client, protobufs, instrumented):

        stash_metadata = json.dumps([instrumented.path, Cardinality.UNARY_STREAM.value])

        response_count = 10

        res = client.unary_stream(
            protobufs.ExampleRequest(
                value="A", delay=10, response_count=response_count
            ),
            timeout=0.05,
            metadata=[("stash", stash_metadata)],
        )
        with pytest.raises(GrpcError) as error:
            list(res)  # client will throw
        assert error.value.code == StatusCode.DEADLINE_EXCEEDED
        assert error.value.message == "Deadline Exceeded"

        time.sleep(0.5)

        # server should not continue to stream responses
        captured_responses = list(instrumented.responses())
        assert len(captured_responses) < response_count

    # add extra test that does mocking and MAKES SURE nameko service is responding
    # correctly (over and above these equivalence tests)


@pytest.mark.equivalence
class TestMethodException:
    def test_error_before_response(self, client, protobufs):
        with pytest.raises(GrpcError) as error:
            client.unary_error(protobufs.ExampleRequest(value="A"))
        assert error.value.code == StatusCode.UNKNOWN
        assert error.value.message == "Exception calling application: boom"

    def test_error_while_streaming_response(self, client, protobufs):
        res = client.stream_error(
            protobufs.ExampleRequest(value="A", response_count=10)
        )
        recvd = []
        with pytest.raises(GrpcError) as error:
            for item in res:
                recvd.append(item)

        # the entrypoint raises after 9 yields, but in the nameko server the exception
        # will abort the stream while some previously read messages are still in
        # the buffer waiting to be sent.
        assert len(recvd) < 10

        assert error.value.code == StatusCode.UNKNOWN
        assert error.value.message == "Exception iterating responses: boom"


class TestRaiseGrpcError:
    @pytest.fixture(params=["server=nameko"])
    def server_type(self, request):
        return request.param[7:]

    def test_error_before_response(self, client, protobufs):
        with pytest.raises(GrpcError) as error:
            client.unary_grpc_error(protobufs.ExampleRequest(value="A"))
        assert error.value.code == StatusCode.UNAUTHENTICATED
        assert error.value.message == "Not allowed!"
        assert isinstance(error.value.status, Status)

    def test_error_while_streaming_response(self, client, protobufs):
        res = client.stream_grpc_error(
            protobufs.ExampleRequest(value="A", response_count=10)
        )
        recvd = []
        with pytest.raises(GrpcError) as error:
            for item in res:
                recvd.append(item)

        # the entrypoint raises after 9 yields, but in the nameko server the exception
        # will abort the stream while some previously read messages are still in
        # the buffer waiting to be sent.
        assert len(recvd) < 10

        assert error.value.code == StatusCode.RESOURCE_EXHAUSTED
        assert error.value.message == "Out of tokens!"
        assert isinstance(error.value.status, Status)


@pytest.mark.equivalence
class TestErrorViaContext:
    def test_error_before_response(self, client, protobufs):
        with pytest.raises(GrpcError) as error:
            client.unary_error_via_context(protobufs.ExampleRequest(value="A"))
        assert error.value.code == StatusCode.UNAUTHENTICATED
        assert error.value.message == "Not allowed!"
        assert isinstance(error.value.status, Status)

    def test_error_while_streaming_response(self, client, protobufs):

        res = client.stream_error_via_context(
            protobufs.ExampleRequest(value="A", response_count=10)
        )
        recvd = []
        with pytest.raises(GrpcError) as error:
            for item in res:
                recvd.append(item)

        # there is no exception raised in the entrypoint here, in contrast to
        # TestMethodException.test_error_while_streaming_response above, so we
        # are guaranteed to receive all 9 messages
        assert len(recvd) == 9

        assert error.value.code == StatusCode.RESOURCE_EXHAUSTED
        assert error.value.message == "Out of tokens!"
        assert isinstance(error.value.status, Status)


class TestErrorDetails:
    @pytest.fixture(params=["client=nameko", "client=dp"])
    def client_type(self, request):
        return request.param[7:]

    @pytest.fixture(params=["server=nameko"])
    def server_type(self, request):
        return request.param[7:]

    @pytest.fixture(params=[True, False])
    def debug_mode(self, request):
        with config.patch({"GRPC_DEBUG": request.param}):
            yield request.param

    def unpack_debug_info(self, detail):
        any_ = Any()
        any_.CopyFrom(detail)
        debug_info = DebugInfo()
        any_.Unpack(debug_info)
        return debug_info

    def test_error_before_response(self, client, protobufs, debug_mode):
        with pytest.raises(GrpcError) as error:
            client.unary_error(protobufs.ExampleRequest(value="A"))
        assert error.value.code == StatusCode.UNKNOWN
        assert error.value.message == "Exception calling application: boom"

        status = error.value.status

        assert status.code == StatusCode.UNKNOWN.value[0]
        assert status.message == "Exception calling application: boom"

        if debug_mode:
            debug_info = self.unpack_debug_info(status.details[0])
            assert debug_info.stack_entries[0] == "Traceback (most recent call last):\n"
            assert debug_info.detail == "boom"
        else:
            assert not status.details

    def test_error_while_streaming_response(self, client, protobufs, debug_mode):
        res = client.stream_error(
            protobufs.ExampleRequest(value="A", response_count=10)
        )
        with pytest.raises(GrpcError) as error:
            list(res)

        assert error.value.code == StatusCode.UNKNOWN
        assert error.value.message == "Exception iterating responses: boom"

        status = error.value.status

        assert status.code == StatusCode.UNKNOWN.value[0]
        assert status.message == "Exception iterating responses: boom"

        if debug_mode:
            debug_info = self.unpack_debug_info(status.details[0])
            assert debug_info.stack_entries[0] == "Traceback (most recent call last):\n"
            assert debug_info.detail == "boom"
        else:
            assert not status.details


class TestCustomErrorFromException:
    @pytest.fixture(params=["server=nameko"])
    def server_type(self, request):
        return request.param[7:]

    @pytest.fixture(autouse=True)
    def register_exception_handler(self):
        from example_nameko import Error

        def handler(exc_info, code=None, message=None):
            exc_type, exc, tb = exc_info

            code = code or StatusCode.PERMISSION_DENIED
            message = "Not allowed!"

            status = Status(
                code=STATUS_CODE_ENUM_TO_INT_MAP[code],
                message=message,
                details=[],  # don't include traceback
            )

            return GrpcError(code=code, message=message, status=status)

        register_exception_handler(Error, handler)

    def test_error_before_response(self, client, protobufs):
        with pytest.raises(GrpcError) as error:
            client.unary_error(protobufs.ExampleRequest(value="A"))
        assert error.value.code == StatusCode.PERMISSION_DENIED
        assert error.value.message == "Not allowed!"

        status = error.value.status

        assert status.code == StatusCode.PERMISSION_DENIED.value[0]
        assert status.message == "Not allowed!"
        assert not status.details

    def test_error_while_streaming_response(self, client, protobufs):
        res = client.stream_error(
            protobufs.ExampleRequest(value="A", response_count=10)
        )
        with pytest.raises(GrpcError) as error:
            list(res)

        assert error.value.code == StatusCode.PERMISSION_DENIED
        assert error.value.message == "Not allowed!"

        status = error.value.status

        assert status.code == StatusCode.PERMISSION_DENIED.value[0]
        assert status.message == "Not allowed!"
        assert not status.details



# -*- coding: utf-8 -*-
from nameko_grpc.entrypoint import Grpc
from nameko_grpc.errors import (
    GRPC_DETAILS_METADATA_KEY,
    GrpcError,
    StatusCode,
    make_status,
)

from helpers import extract_metadata, instrumented, maybe_echo_metadata, maybe_sleep

import example_pb2_grpc
from example_pb2 import ExampleReply


class Error(Exception):
    pass


grpc = Grpc.implementing(example_pb2_grpc.exampleStub)


class example:
    name = "example"

    @grpc
    @instrumented
    def unary_unary(self, request, context):
        metadata = extract_metadata(context)
        maybe_echo_metadata(context)
        maybe_sleep(request)
        message = request.value * (request.multiplier or 1)
        return ExampleReply(message=message, metadata=metadata)

    @grpc
    @instrumented
    def unary_stream(self, request, context):
        metadata = extract_metadata(context)
        maybe_echo_metadata(context)
        message = request.value * (request.multiplier or 1)
        for i in range(request.response_count):
            maybe_sleep(request)
            yield ExampleReply(message=message, seqno=i + 1, metadata=metadata)

    @grpc
    @instrumented
    def stream_unary(self, request, context):
        metadata = extract_metadata(context)
        maybe_echo_metadata(context)
        messages = []
        for index, req in enumerate(request):
            maybe_sleep(req)
            message = req.value * (req.multiplier or 1)
            messages.append(message)

        return ExampleReply(message=",".join(messages), metadata=metadata)

    @grpc
    @instrumented
    def stream_stream(self, request, context):
        metadata = extract_metadata(context)
        maybe_echo_metadata(context)
        for index, req in enumerate(request):
            maybe_sleep(req)
            message = req.value * (req.multiplier or 1)
            yield ExampleReply(message=message, seqno=index + 1, metadata=metadata)

    @grpc(expected_exceptions=Error)
    @instrumented
    def unary_error(self, request, context):
        maybe_echo_metadata(context)
        maybe_sleep(request)
        raise Error("boom")

    @grpc
    @instrumented
    def unary_error_via_context(self, request, context):
        code = StatusCode.UNAUTHENTICATED
        message = "Not allowed!"

        context.set_code(code)
        context.set_message(message)
        context.set_trailing_metadata(
            [
                (
                    GRPC_DETAILS_METADATA_KEY,
                    make_status(code, message).SerializeToString(),
                )
            ]
        )

    @grpc(expected_exceptions=Error)
    @instrumented
    def unary_grpc_error(self, request, context):
        maybe_echo_metadata(context)
        maybe_sleep(request)

        code = StatusCode.UNAUTHENTICATED
        message = "Not allowed!"

        raise GrpcError(code=code, message=message, status=make_status(code, message))

    @grpc(expected_exceptions=Error)
    @instrumented
    def stream_error(self, request, context):
        metadata = extract_metadata(context)
        maybe_echo_metadata(context)
        message = request.value * (request.multiplier or 1)
        for i in range(request.response_count):
            maybe_sleep(request)
            # raise on the last message
            if i == request.response_count - 1:
                raise Error("boom")
            yield ExampleReply(message=message, seqno=i + 1, metadata=metadata)

    @grpc
    @instrumented
    def stream_error_via_context(self, request, context):
        message = request.value * (request.multiplier or 1)
        for i in range(request.response_count):
            maybe_sleep(request)
            # break on the last message
            if i == request.response_count - 1:

                code = StatusCode.RESOURCE_EXHAUSTED
                message = "Out of tokens!"

                context.set_code(code)
                context.set_message(message)
                context.set_trailing_metadata(
                    [
                        (
                            GRPC_DETAILS_METADATA_KEY,
                            make_status(code, message).SerializeToString(),
                        )
                    ]
                )
                break
            yield ExampleReply(message=message, seqno=i + 1)

    @grpc(expected_exceptions=Error)
    @instrumented
    def stream_grpc_error(self, request, context):
        metadata = extract_metadata(context)
        maybe_echo_metadata(context)
        message = request.value * (request.multiplier or 1)
        for i in range(request.response_count):
            maybe_sleep(request)
            # raise on the last message
            if i == request.response_count - 1:

                code = StatusCode.RESOURCE_EXHAUSTED
                message = "Out of tokens!"

                raise GrpcError(
                    code=code, message=message, status=make_status(code, message),
                )
            yield ExampleReply(message=message, seqno=i + 1, metadata=metadata)



# -*- coding: utf-8 -*-
import gzip
import zlib

import pytest
from grpc import StatusCode
from mock import patch

from nameko_grpc.compression import (
    UnsupportedEncoding,
    compress,
    decompress,
    select_algorithm,
)
from nameko_grpc.errors import GrpcError


@pytest.mark.equivalence
class TestCompression:
    """ Some of these tests are incomplete because the standard Python GRPC client
    and/or server don't conform to the spec.

    See https://stackoverflow.com/questions/53233339/what-is-the-state-of-compression-
    for more details.
    """

    @pytest.fixture(params=["deflate", "gzip", "none"])
    def compression_algorithm(self, request):
        return request.param

    @pytest.fixture(params=["high", "medium", "low", "none"])
    def compression_level(self, request):
        return request.param

    @pytest.fixture
    def server(self, start_server, compression_algorithm, compression_level):
        return start_server(
            "example",
            compression_algorithm=compression_algorithm,
            compression_level=compression_level,
        )

    @pytest.fixture
    def client(self, start_client, compression_algorithm, compression_level, server):
        return start_client(
            "example",
            compression_algorithm=compression_algorithm,
            compression_level=compression_level,
        )

    def test_default_compression(self, client, protobufs):

        response = client.unary_unary(protobufs.ExampleRequest(value="A"))
        assert response.message == "A"

        responses = client.unary_stream(
            protobufs.ExampleRequest(value="A", response_count=2)
        )
        assert [(response.message, response.seqno) for response in responses] == [
            ("A", 1),
            ("A", 2),
        ]

        def generate_requests():
            for value in ["A", "B"]:
                yield protobufs.ExampleRequest(value=value)

        response = client.stream_unary(generate_requests())
        assert response.message == "A,B"

        def generate_requests():
            for value in ["A", "B"]:
                yield protobufs.ExampleRequest(value=value)

        responses = client.stream_stream(generate_requests())
        assert [(response.message, response.seqno) for response in responses] == [
            ("A", 1),
            ("B", 2),
        ]

    @pytest.mark.parametrize("algorithm_for_call", ["gzip", "identity"])
    def test_set_different_compression_for_call(
        self, start_client, start_server, protobufs, algorithm_for_call
    ):
        start_server("example")
        client = start_client("example", compression_algorithm="deflate")

        response = client.unary_unary(
            protobufs.ExampleRequest(value="A" * 1000), compression=algorithm_for_call
        )
        assert response.message == "A" * 1000

        responses = client.unary_stream(
            protobufs.ExampleRequest(value="A" * 1000, response_count=2),
            compression=algorithm_for_call,
        )
        assert [(response.message, response.seqno) for response in responses] == [
            ("A" * 1000, 1),
            ("A" * 1000, 2),
        ]

        def generate_requests():
            for value in ["A" * 1000, "B" * 1000]:
                yield protobufs.ExampleRequest(value=value)

        response = client.stream_unary(
            generate_requests(), compression=algorithm_for_call
        )
        assert response.message == "A" * 1000 + "," + "B" * 1000

        def generate_requests():
            for value in ["A" * 1000, "B" * 1000]:
                yield protobufs.ExampleRequest(value=value)

        responses = client.stream_stream(
            generate_requests(), compression=algorithm_for_call
        )
        assert [(response.message, response.seqno) for response in responses] == [
            ("A" * 1000, 1),
            ("B" * 1000, 2),
        ]

    def test_request_unsupported_algorithm_at_client(
        self, start_client, start_server, protobufs, client_type, server_type
    ):
        if client_type == "grpc":
            pytest.skip("grpc client throws assertion error")

        start_server("example")
        client = start_client("example")

        response = client.unary_unary(
            protobufs.ExampleRequest(value="A"), compression="bogus"
        )
        assert response.message == "A"

        responses = client.unary_stream(
            protobufs.ExampleRequest(value="A", response_count=2), compression="bogus"
        )
        assert [(response.message, response.seqno) for response in responses] == [
            ("A", 1),
            ("A", 2),
        ]

        def generate_requests():
            for value in ["A", "B"]:
                yield protobufs.ExampleRequest(value=value)

        response = client.stream_unary(generate_requests(), compression="bogus")
        assert response.message == "A,B"

        def generate_requests():
            for value in ["A", "B"]:
                yield protobufs.ExampleRequest(value=value)

        responses = client.stream_stream(generate_requests(), compression="bogus")
        assert [(response.message, response.seqno) for response in responses] == [
            ("A", 1),
            ("B", 2),
        ]

        # TODO assert that the message is sent _without_ compression

    @patch("nameko_grpc.entrypoint.SUPPORTED_ENCODINGS", new=["deflate"])
    def test_request_unsupported_algorithm_at_server(
        self, start_client, start_server, protobufs, client_type, server_type
    ):
        """ It's not possible to run this test with the GRPC server.

        The server can be configured to disable certain compression algorithms with
        the grpc.compression_enabled_algorithms_bitset server option. Unfortunately,
        making a call with a disabled algorithm results in an unhandled exception
        rather than the error described in the spec.
        """

        if server_type == "grpc":
            pytest.skip("See docstring for details")

        start_server("example")
        client = start_client("example")

        with pytest.raises(GrpcError) as error:
            client.unary_unary(
                protobufs.ExampleRequest(value="A" * 1000), compression="gzip"
            )
        assert error.value.code == StatusCode.UNIMPLEMENTED
        assert error.value.message == "Algorithm not supported: gzip"

        res = client.unary_stream(
            protobufs.ExampleRequest(value="A" * 1000, response_count=2),
            compression="gzip",
        )
        with pytest.raises(GrpcError) as error:
            list(res)
        assert error.value.code == StatusCode.UNIMPLEMENTED
        assert error.value.message == "Algorithm not supported: gzip"

        def generate_requests():
            for value in ["A" * 1000, "B" * 1000]:
                yield protobufs.ExampleRequest(value=value)

        with pytest.raises(GrpcError) as error:
            client.stream_unary(generate_requests(), compression="gzip")
        assert error.value.code == StatusCode.UNIMPLEMENTED
        assert error.value.message == "Algorithm not supported: gzip"

        def generate_requests():
            for value in ["A" * 1000, "B" * 1000]:
                yield protobufs.ExampleRequest(value=value)

        res = client.stream_stream(generate_requests(), compression="gzip")
        with pytest.raises(GrpcError) as error:
            next(res)
        assert error.value.code == StatusCode.UNIMPLEMENTED
        assert error.value.message == "Algorithm not supported: gzip"

    def test_compression_not_required_at_client(self):
        # TODO
        pass

    def test_compression_not_required_at_server(self):
        # TODO
        pass

    def test_respond_with_different_algorithm(self):
        """ The GRPC server doesn't seem to support this behaviour. It's either
        explicitly not supported, or there is a bug that results in an error if the
        server also has a default compression algorithm set.

        The Nameko server MAY be able to support this behaviour in future, once it's
        clear from the reference implementation exactly how it should behave.

        """
        pytest.skip("See docstring for details")

    def test_disable_compression_for_message(self):
        """ The non-beta GRPC client doesn't support any way to disable compression
        for a call. Additionally, it's not clear how the client should request that
        an the next _message_ should be sent uncompressed as described in the spec.

        It is possible to call context.disable_next_message_compression() in a GRPC
        server method, but this results in an "Unallowed duplicate metadata" error.

        The Nameko server MAY be able to support this behaviour in future, once it's
        clear from the reference implementation exactly how it should behave.
        """
        pytest.skip("See docstring for details")


class TestDecompress:
    def test_preferred_algorithm(self):
        payload = b"\x00" * 1000

        assert decompress(zlib.compress(payload)) == payload

    def test_fallback_algorithm(self):
        payload = b"\x00" * 1000

        assert decompress(gzip.compress(payload)) == payload

    def test_unsupported_algorithm(self):
        payload = b"\x00" * 1000

        with pytest.raises(UnsupportedEncoding):
            decompress(payload)  # not correctly encoded


class TestCompress:
    def test_identity(self):
        payload = b"\x00" * 1000
        assert compress(payload, "identity") == (False, payload)

    def test_deflate(self):
        payload = b"\x00" * 1000
        assert compress(payload, "deflate") == (True, zlib.compress(payload))

    def test_gzip(self):
        payload = b"\x00" * 1000
        assert compress(payload, "gzip") == (True, gzip.compress(payload))

    def test_unsupported_algorithm(self):
        payload = b"\x00" * 1000

        with pytest.raises(UnsupportedEncoding):
            compress(payload, "bogus")


class TestSelectAlgorithm:
    @pytest.mark.parametrize("preferred", ["deflate", "gzip", "identity"])
    def test_preferred_algorithm_available(self, preferred):
        acceptable = "deflate,gzip,identity"
        expected = preferred
        assert select_algorithm(acceptable, preferred) == expected

    def test_preferred_algorithm_is_null(self):
        acceptable = "deflate,gzip,identity"
        preferred = None
        expected = "deflate"
        assert select_algorithm(acceptable, preferred) == expected

    def test_fallback_to_supported_algorithm(self):
        acceptable = "minify,gzip,identity"
        preferred = "minify"
        expected = "gzip"
        assert select_algorithm(acceptable, preferred) == expected

    def test_fallback_to_identity(self):
        acceptable = "minify,identity"
        preferred = "minify"
        expected = "identity"
        assert select_algorithm(acceptable, preferred) == expected

    def test_no_supported_algo(self):
        acceptable = "minify"
        preferred = "minify"
        with pytest.raises(UnsupportedEncoding):
            select_algorithm(acceptable, preferred)



# -*- coding: utf-8 -*-
import threading
import time
from logging import getLogger
from urllib.parse import urlparse

from grpc import StatusCode

from nameko_grpc.channel import ClientChannel
from nameko_grpc.compression import SUPPORTED_ENCODINGS
from nameko_grpc.constants import Cardinality
from nameko_grpc.errors import GrpcError
from nameko_grpc.inspection import Inspector
from nameko_grpc.ssl import SslConfig
from nameko_grpc.timeout import bucket_timeout


log = getLogger(__name__)


USER_AGENT = "grpc-python-nameko/0.0.1"
CONTENT_TYPE = "application/grpc+proto"


class Future:
    def __init__(self, response_stream, output_type, cardinality):
        self.response_stream = response_stream
        self.output_type = output_type
        self.cardinality = cardinality

    def initial_metadata(self):
        return self.response_stream.headers.for_application

    def trailing_metadata(self):
        return self.response_stream.trailers.for_application

    def result(self):
        response = self.response_stream.consume(self.output_type)
        if self.cardinality in (Cardinality.STREAM_UNARY, Cardinality.UNARY_UNARY):
            response = next(response)
        return response


class Method:
    def __init__(self, client, name, extra_metadata=None):
        self.client = client
        self.name = name
        self.extra_metadata = extra_metadata or []

    def __call__(self, request, **kwargs):
        return self.future(request, **kwargs).result()

    def future(self, request, timeout=None, compression=None, metadata=None):
        inspector = Inspector(self.client.stub)

        cardinality = inspector.cardinality_for_method(self.name)
        input_type = inspector.input_type_for_method(self.name)
        output_type = inspector.output_type_for_method(self.name)
        service_name = inspector.service_name

        compression = compression or self.client.default_compression
        if compression not in SUPPORTED_ENCODINGS:
            log.warning(
                "Invalid compression algorithm: '{}'. Ignoring.".format(compression)
            )
            compression = self.client.default_compression

        scheme = "https" if self.client.ssl else "http"

        request_headers = [
            (":method", "POST"),
            (":scheme", scheme),
            (":authority", urlparse(self.client.target).hostname),
            (":path", "/{}/{}".format(inspector.service_name, self.name)),
            ("te", "trailers"),
            ("content-type", CONTENT_TYPE),
            ("user-agent", USER_AGENT),
            ("grpc-encoding", compression),
            ("grpc-message-type", "{}.{}".format(service_name, input_type.__name__)),
            ("grpc-accept-encoding", ",".join(SUPPORTED_ENCODINGS)),
        ]

        if metadata is not None:
            metadata = metadata[:]
        else:
            metadata = []

        metadata.extend(self.extra_metadata)

        for key, value in metadata:
            request_headers.append((key, value))

        if timeout is not None:
            request_headers.append(("grpc-timeout", bucket_timeout(timeout)))

        if cardinality in (Cardinality.UNARY_UNARY, Cardinality.UNARY_STREAM):
            request = (request,)

        response_stream = self.client.invoke(request_headers, request, timeout)

        return Future(response_stream, output_type, cardinality)


class Proxy:
    def __init__(self, client):
        self.client = client

    def __getattr__(self, name):
        return Method(self.client, name)


class ClientBase:

    manager = None
    sock = None

    def __init__(
        self,
        target,
        stub,
        compression_algorithm="none",
        compression_level="high",
        ssl=False,
    ):
        self.target = target
        self.stub = stub
        self.compression_algorithm = compression_algorithm
        self.compression_level = compression_level  # NOTE not used
        self.ssl = SslConfig(ssl)

    def spawn_thread(self, target, args=(), kwargs=None, name=None):
        raise NotImplementedError

    @property
    def default_compression(self):
        if self.compression_algorithm != "none":
            return self.compression_algorithm
        return "identity"

    def start(self):
        self.channel = ClientChannel(self.target, self.ssl, self.spawn_thread)
        self.channel.start()

    def stop(self):
        self.channel.stop()

    def timeout(self, send_stream, response_stream, deadline):
        start = time.time()
        while True:
            if send_stream.closed and response_stream.closed:
                break
            elapsed = time.time() - start
            if elapsed > deadline:
                error = GrpcError(
                    code=StatusCode.DEADLINE_EXCEEDED, message="Deadline Exceeded"
                )
                response_stream.close(error)
                send_stream.close()
                break
            time.sleep(0.001)

    def invoke(self, request_headers, request, timeout):
        send_stream, response_stream = self.channel.send_request(request_headers)
        if timeout:
            self.spawn_thread(
                target=self.timeout,
                args=(send_stream, response_stream, timeout),
                name=f"client timeout [{request}]",
            )
        self.spawn_thread(
            target=send_stream.populate,
            args=(request,),
            name=f"populate request [{request}]",
        )
        return response_stream


class Client(ClientBase):
    """ Standalone gRPC client that uses native threads.
    """

    def __enter__(self):
        return self.start()

    def __exit__(self, *args):
        self.stop()

    def start(self):
        super().start()
        return Proxy(self)

    def spawn_thread(self, target, args=(), kwargs=None, name=None):
        threading.Thread(target=target, args=args, kwargs=kwargs, name=name).start()



# -*- coding: utf-8 -*-
import os
import socket
import subprocess
import sys
import threading
import time
import uuid
from importlib import import_module
from unittest.mock import patch

import pytest
from eventlet.green import zmq
from mock import Mock
from nameko import config
from nameko.testing.services import dummy
from nameko.testing.utils import find_free_port, get_extension

from nameko_grpc.client import Client
from nameko_grpc.dependency_provider import GrpcProxy
from nameko_grpc.inspection import Inspector

from helpers import Command, RemoteClientTransport, Stash


def pytest_addoption(parser):

    parser.addoption(
        "--client",
        action="store",
        choices=["nameko", "dp", "grpc", "all"],
        dest="client",
        default="all",
        help="Use this client type",
    )

    parser.addoption(
        "--server",
        action="store",
        choices=["nameko", "grpc", "all"],
        dest="server",
        default="all",
        help="Use this server type",
    )


@pytest.fixture(scope="session")
def spec_dir(tmpdir_factory):
    master = os.path.join(os.path.dirname(__file__), "spec")
    temp = tmpdir_factory.mktemp("spec")
    for filename in os.listdir(master):
        path = os.path.join(master, filename)
        if os.path.isfile(path):
            copy = temp.join(filename)
            with open(path) as file:
                copy.write(file.read())

    sys.path.append(temp.strpath)
    yield temp
    sys.path.remove(temp.strpath)


@pytest.fixture(scope="session")
def compile_proto(spec_dir):

    spec_path = spec_dir.strpath

    def codegen(proto_name):

        proto_path = os.path.join(spec_path, "{}.proto".format(proto_name))
        proto_last_modified = os.path.getmtime(proto_path)

        for generated_file in (
            "{}_pb2.py".format(proto_name),
            "{}_pb2_grpc.py".format(proto_name),
        ):
            generated_path = os.path.join(spec_path, generated_file)
            if (
                not os.path.exists(generated_path)
                or os.path.getmtime(generated_path) < proto_last_modified
            ):
                protoc_args = [
                    "-I{}".format(spec_path),
                    "--python_out",
                    spec_path,
                    "--grpc_python_out",
                    spec_path,
                    proto_path,
                ]
                # protoc.main is confused by absolute paths, so use subprocess instead
                python_args = ["python", "-m", "grpc_tools.protoc"] + protoc_args
                subprocess.call(python_args)

        protobufs = import_module("{}_pb2".format(proto_name))
        stubs = import_module("{}_pb2_grpc".format(proto_name))

        return protobufs, stubs

    return codegen


@pytest.fixture(scope="session")
def load_protobufs(compile_proto):
    def load(name):
        protobufs, _ = compile_proto(name)
        return protobufs

    return load


@pytest.fixture(scope="session")
def load_stubs(compile_proto):
    def load(name):
        _, stubs = compile_proto(name)
        return stubs

    return load


@pytest.fixture(autouse=True, scope="session")
def example_proto(compile_proto):
    compile_proto("example")


@pytest.fixture
def spawn_process():

    procs = []

    def spawn(*args, env=None):
        popen_args = [sys.executable]
        popen_args.extend(args)
        procs.append(subprocess.Popen(popen_args, env=env))

    yield spawn

    for proc in procs:
        proc.terminate()


@pytest.fixture
def grpc_port():
    return find_free_port()


@pytest.fixture
def start_grpc_server(request, compile_proto, spawn_process, spec_dir, grpc_port):

    secure = "secure" if request.node.get_closest_marker(name="secure") else "insecure"

    server_script = os.path.join(os.path.dirname(__file__), "grpc_indirect_server.py")

    def make(
        service_name,
        proto_name=None,
        compression_algorithm="none",
        compression_level="high",
    ):
        if proto_name is None:
            proto_name = service_name

        env = os.environ.copy()
        env["PYTHONPATH"] = spec_dir.strpath

        spawn_process(
            server_script,
            str(grpc_port),
            secure,
            proto_name,
            service_name,
            compression_algorithm,
            compression_level,
            env=env,
        )

        # wait for server to start
        while True:
            try:
                sock = socket.socket()
                sock.connect(("127.0.0.1", grpc_port))
                sock.close()
                break
            except socket.error:
                time.sleep(0.1)

    yield make


@pytest.fixture
def start_grpc_client(request, load_stubs, spawn_process, spec_dir, grpc_port):

    secure = "secure" if request.node.get_closest_marker(name="secure") else "insecure"

    client_script = os.path.join(os.path.dirname(__file__), "grpc_indirect_client.py")

    clients = []

    context = zmq.Context()

    class Result:
        _metadata = None

        def __init__(self, command):
            self.command = command

        @property
        def metadata(self):
            if self._metadata is None:
                self._metadata = self.command.get_metadata()
            return self._metadata

        def code(self):
            return self.metadata.get("code")

        def details(self):
            return self.metadata.get("details")

        def initial_metadata(self):
            return self.metadata.get("initial_metadata")

        def trailing_metadata(self):
            return self.metadata.get("trailing_metadata")

        def result(self):
            return self.command.get_response()

    class Method:
        def __init__(self, client, name):
            self.client = client
            self.name = name

        def __call__(self, request, **kwargs):
            return self.future(request, **kwargs).result()

        def future(self, request, **kwargs):
            inspector = Inspector(self.client.stub)

            cardinality = inspector.cardinality_for_method(self.name)

            command = Command(self.name, cardinality, kwargs, self.client.transport)
            command.issue()
            threading.Thread(target=command.send_request, args=(request,)).start()
            return Result(command)

    class Client:
        def __init__(self, stub, transport):
            self.stub = stub
            self.transport = transport

        def __getattr__(self, name):
            return Method(self, name)

        def shutdown(self):
            self.transport.send(Command.END, close=True)

    def make(
        service_name,
        proto_name=None,
        compression_algorithm="none",
        compression_level="high",
    ):
        if proto_name is None:
            proto_name = service_name

        stubs = load_stubs(proto_name)
        stub_cls = getattr(stubs, "{}Stub".format(service_name))

        transport, zmq_port = RemoteClientTransport.bind_to_free_port(
            context, zmq.REQ, "tcp://127.0.0.1"
        )

        env = os.environ.copy()
        env["PYTHONPATH"] = spec_dir.strpath

        spawn_process(
            client_script,
            str(grpc_port),
            secure,
            proto_name,
            service_name,
            compression_algorithm,
            compression_level,
            str(zmq_port),
            env=env,
        )

        client = Client(stub_cls, transport)
        clients.append(client)
        return client

    yield make

    # shut down indirect clients
    for client in clients:
        client.shutdown()


@pytest.fixture
def start_nameko_server(request, spec_dir, container_factory, grpc_port):

    if request.node.get_closest_marker(name="secure"):
        ssl_options = {
            "cert_chain": {
                "keyfile": "test/certs/server.key",
                "certfile": "test/certs/server.crt",
            }
        }
    else:
        ssl_options = False

    def make(
        service_name,
        proto_name=None,
        compression_algorithm="none",
        compression_level="high",
    ):
        if proto_name is None:
            proto_name = service_name

        service_module = import_module("{}_nameko".format(proto_name))
        service_cls = getattr(service_module, service_name)

        conf = {
            "GRPC_BIND_PORT": grpc_port,
            "GRPC_COMPRESSION_ALGORITHM": compression_algorithm,
            "GRPC_COMPRESSION_LEVEL": compression_level,
        }
        if ssl_options:
            conf.update({"GRPC_SSL": ssl_options})

        config.setup(conf)
        container = container_factory(service_cls)
        container.start()

        return container

    yield make


@pytest.fixture
def start_nameko_client(request, load_stubs, spec_dir, grpc_port):

    clients = []

    if request.node.get_closest_marker(name="secure"):
        ssl_options = {"verify_mode": "none", "check_hostname": False}
    else:
        ssl_options = False

    def make(
        service_name,
        proto_name=None,
        compression_algorithm="none",
        compression_level="high",
    ):
        if proto_name is None:
            proto_name = service_name

        stubs = load_stubs(proto_name)
        stub_cls = getattr(stubs, "{}Stub".format(service_name))
        client = Client(
            "//localhost:{}".format(grpc_port),
            stub_cls,
            compression_algorithm,
            compression_level,
            ssl_options,
        )
        clients.append(client)
        return client.start()

    yield make

    for client in clients:
        client.stop()


@pytest.fixture
def start_dependency_provider(
    request, load_stubs, spec_dir, grpc_port, container_factory
):

    clients = []

    if request.node.get_closest_marker(name="secure"):
        ssl_config = {"verify_mode": "none", "check_hostname": False}
    else:
        ssl_config = False

    # TODO turn into config.module_patch?
    patch_config = config.copy()
    patch_config["GRPC_SSL"] = ssl_config
    with patch("nameko_grpc.dependency_provider.config", new=patch_config):

        def make(
            service_name,
            proto_name=None,
            compression_algorithm="none",
            compression_level="high",
        ):
            if proto_name is None:
                proto_name = service_name

            stubs = load_stubs(proto_name)
            stub_cls = getattr(stubs, "{}Stub".format(service_name))

            class Service:
                name = "caller"

                example_grpc = GrpcProxy(
                    "//localhost:{}".format(grpc_port),
                    stub_cls,
                    compression_algorithm=compression_algorithm,
                    compression_level=compression_level,
                )

                @dummy
                def call(self):
                    pass

            container = container_factory(Service)
            container.start()

            grpc_proxy = get_extension(container, GrpcProxy)
            clients.append(grpc_proxy)
            return grpc_proxy.get_dependency(Mock(context_data={}))

        yield make

    for client in clients:
        client.stop()


@pytest.fixture(params=["server=grpc", "server=nameko"])
def server_type(request):
    return request.param[7:]


@pytest.fixture
def start_server(request, server_type):
    if server_type == "grpc":
        if request.config.option.server not in ("grpc", "all"):
            pytest.skip("grpc server not requested")
        return request.getfixturevalue("start_grpc_server")
    if server_type == "nameko":
        if request.config.option.server not in ("nameko", "all"):
            pytest.skip("nameko server not requested")
        return request.getfixturevalue("start_nameko_server")


@pytest.fixture(params=["client=grpc", "client=nameko", "client=dp"])
def client_type(request):
    return request.param[7:]


@pytest.fixture
def start_client(request, client_type, start_server):
    if client_type == "grpc":
        if request.config.option.client not in ("grpc", "all"):
            pytest.skip("grpc client not requested")
        return request.getfixturevalue("start_grpc_client")
    if client_type == "nameko":
        if request.config.option.client not in ("nameko", "all"):
            pytest.skip("nameko client not requested")
        return request.getfixturevalue("start_nameko_client")
    if client_type == "dp":
        if request.config.option.client not in ("dp", "all"):
            pytest.skip("dp client not requested")
        return request.getfixturevalue("start_dependency_provider")


@pytest.fixture
def instrumented(tmpdir_factory):
    stashes = tmpdir_factory.mktemp("instrument_stashes")
    stash_file = stashes.join(str(uuid.uuid4()))
    return Stash(stash_file.strpath)


@pytest.fixture
def server(start_server):
    return start_server("example")


@pytest.fixture
def client(start_client, server):
    return start_client("example")


@pytest.fixture
def stubs(load_stubs):
    return load_stubs("example")


@pytest.fixture
def protobufs(load_protobufs):
    return load_protobufs("example")
